/** @file fifo_media.cpp
 * Mediator of FIFO reference model.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <fifo_media.h>
#include <fifo_interface.h>

#include <iostream>

namespace cpptesk {
namespace examples {
namespace fifo {

FIFOMediator::FIFOMediator()
{
    // Registration of serializers, deserializers to interfaces.
    CPPTESK_SET_INPUT_ADAPTER(iface1, FIFOMediator::serialize_iface1);
    CPPTESK_SET_INPUT_ADAPTER(iface2, FIFOMediator::serialize_iface2);
    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface3, FIFOMediator::deserialize_iface3);

    // Setup of maximum time of waiting.
    CPPTESK_SET_REACTION_TIMEOUT(FIFOMediator::REACTION_TIMEOUT);

    // Initialization of structures to keep signals from DUT.
    clear_fifo_inputs(&inputs);
    clear_fifo_outputs(&outputs);
}

FIFOMediator::~FIFOMediator() {}

bool FIFOMediator::isIface1Ready() const
{
    return FIFO::isIface1Ready();
}
bool FIFOMediator::isIface2Ready() const
{
    return FIFO::isIface2Ready();
}

CPPTESK_DEFINE_PROCESS(FIFOMediator::serialize_iface1)
{
    // Copy data into local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start working.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "FIFOMediator::serialize_iface1 " << msg);

    // Capture interface iface1.
    CPPTESK_CAPTURE_IFACE();

    // Set data to wires of DUT.
    inputs.CE_WR = 1;
    inputs.DI = msg.get_data();

    // Release interface iface1.
    CPPTESK_RELEASE_IFACE();

    // Wait for 1 cycle of simulation.
    CPPTESK_CYCLE();

    // Stop processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(FIFOMediator::serialize_iface2)
{
    // Copy data into local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start working.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "FIFOMediator::serialize_iface2 " << msg);

    // Capture interface iface2.
    CPPTESK_CAPTURE_IFACE();

    // Set data to wires of DUT.
    inputs.CE_RD = 1;

    // Release interface iface2.
    CPPTESK_RELEASE_IFACE();

    // Wait for 1 cycle of simulation.
    CPPTESK_CYCLE();

    // Stop processing.
    CPPTESK_STOP_PROCESS();
}

CPPTESK_DEFINE_PROCESS(FIFOMediator::deserialize_iface3)
{
    // Copy data into local variable.
    OutputData &msg = CPPTESK_CAST_MESSAGE(OutputData);

    // Start working.
    CPPTESK_START_PROCESS();

    CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_USER, "FIFOMediator::deserialize_iface3 " << msg.forceToString());

    // Wait for DO_VAL signal.
    CPPTESK_WAIT_REACTION(outputs.DO_VAL);

    // Received data should be saved.
    msg.set_data(outputs.DO);

    // Jump to the following reaction if it is given.
    CPPTESK_NEXT_REACTION();

    // Stop processing.
    CPPTESK_STOP_PROCESS();
}

void FIFOMediator::initialize()
{
//    ::start_testbench();
}

void FIFOMediator::finalize()
{
//    ::stop_testbench();
}

void FIFOMediator::setInputs()
{
    ::map_fifo_state_down(&inputs);
}

void FIFOMediator::getOutputs()
{
    ::map_fifo_state_up(&outputs);
}

void FIFOMediator::simulate()
{
}

}}}

