/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Generalized model of a hardware component
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/debug.hpp>

#include <assert.h>
#include <string>

#include <hw/process.hpp>
#include <hw/time.hpp>

/** @def CPPTESK_COMPONENT
 * Starts user component declaration.
 */
#define CPPTESK_COMPONENT(UserComponent) \
    struct UserComponent: public ::cpptesk::hw::Component

/** @def CPPTESK_CAPTURE_COMPONENT
 * Captures a component.
 */
#define CPPTESK_CAPTURE_COMPONENT(component) \
    do { \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "[CAPTURE] (component=" << component.getName() << ")"); \
        CPPTESK_ASSERTION(((component).getStatus() == ::cpptesk::hw::Component::LOCK || (component).getStatus() == ::cpptesk::hw::Component::IDLE), \
            "Component " << (component).getName() << " has been already captured"); \
        (component).capture(); \
    } while(false)

/** @def CPPTESK_RELEASE_COMPONENT
 * Releases a component.
 */
#define CPPTESK_RELEASE_COMPONENT(component) \
    do { \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "[RELEASE] (component=" << component.getName() << ")"); \
        CPPTESK_ASSERTION(((component).getStatus() == ::cpptesk::hw::Component::BUSY || (component).getStatus() == ::cpptesk::hw::Component::LOCK), \
            "Component " << (component).getName() << " has been already released"); \
        (component).release(); \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define COMPONENT         CPPTESK_COMPONENT 
    #define CAPTURE_COMPONENT CPPTESK_CAPTURE_COMPONENT
    #define RELEASE_COMPONENT CPPTESK_RELEASE_COMPONENT
#endif // HW_SHORT_NAMES

#define __CPPTESK_DEFAULT_COMPONENT_NAME "<unregistered>"

namespace cpptesk {
namespace hw {

class Component 
{
public:
    enum status_t { IDLE, LOCK, BUSY };

    Component(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME):
        _status(IDLE), _name(name), _timer(0), _parent(0) {}

    virtual ~Component() {}

    const std::string& getName() const { return _name; }
    void setName(const std::string &name) { _name = name; }

    void updateName(const std::string &name)
    {
        if(_name == __CPPTESK_DEFAULT_COMPONENT_NAME)
            { _name = name; }
    }

    virtual void request(const Process &process)
    {
        CPPTESK_ASSERTION(_status == IDLE,
            "Component " << _name << " has been already requested");
        _status = LOCK; _owner = &process;
    }

    virtual void capture()
    {
        CPPTESK_ASSERTION((_status == LOCK || _status == IDLE),
            "Component " << _name << " has been already captured");
        _status = BUSY;
    }

    virtual void release()
    {
        CPPTESK_ASSERTION((_status == BUSY || _status == LOCK),
            "Component " << _name << " has been already released");
        _status = IDLE; _timer = 0; _owner = 0;
    }

    virtual void cycle()
    {
        if(_status == BUSY)
            { _timer++; }
    }

    status_t getStatus() const { return _status; }

    bool isReady() const { return _status == IDLE; }
    bool isBusy() const { return !isReady(); }

    const Process* getOwner() const { return _owner; }

    cycle_t getTimer() const { return _timer; }

    void setParent(const void *parent) { _parent = parent; }
    const void* getParent() const { return _parent; }

protected:
    /// Current owner of the component.
    const Process *_owner;
    /// Status of the component.
    status_t _status;
    /// Name of the component.
    std::string _name;
    /// Usage time.
    cycle_t _timer;
    /// Parent (a model containing the component).
    const void *_parent;
};

}} // namespace cpptesk::hw

