/*Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Associative Elimination 
//           (Rule # 3 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[a:b]+[b:a] => [a:a]+[b:b]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/transform_unit.hpp>
#include <hw/diagnostics/rule/associative.hpp>

#include <stdio.h>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet AssociativeEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_1, ::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_2)    // Collapse the reactions
    {
        ReactionSet _transform_out;
        
        // Message from reaction 1
        const ::cpptesk::hw::Message *_message1Spec = _reactionUnit_1._reaction.getSpecMessage();
        const ::cpptesk::hw::Message *_message1Impl = _reactionUnit_1._reaction.getImplMessage();
        // Message from reaction 2
        const ::cpptesk::hw::Message *_message2Spec = _reactionUnit_2._reaction.getSpecMessage();
        const ::cpptesk::hw::Message *_message2Impl = _reactionUnit_2._reaction.getImplMessage();
        
        // Get "start" and "end" time of reactions
        cycle_t start_time   =  _reactionUnit_1._reaction.whenStarted();
        cycle_t start_time_2 =  _reactionUnit_2._reaction.whenStarted();
        cycle_t end_time     =  _reactionUnit_1._reaction.whenStopped();
        cycle_t end_time_2   =  _reactionUnit_2._reaction.whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        const ::cpptesk::hw::Interface *iface1 = _reactionUnit_1._reaction.where();
        const ::cpptesk::hw::Interface *iface2 = _reactionUnit_2._reaction.where();
        
        ::cpptesk::hw::Reaction reactionNormal_1 = Reaction(NORMAL_REACTION, *iface1, *_message1Spec, *_message2Impl, start_time, end_time, "normal reaction from rule#4");
        ::cpptesk::hw::Reaction reactionNormal_2 = Reaction(NORMAL_REACTION, *iface2, *_message2Spec, *_message1Impl, start_time, end_time, "normal reaction from rule#4");
        
        _transform_out.addReaction(reactionNormal_1);
        _transform_out.addReaction(reactionNormal_2);
        
        // Set flag
        _reactionUnit_1.useInCollapse();
        _reactionUnit_2.useInCollapse();
        
        return _transform_out;
    }

    int AssociativeEliminationRule::match(::cpptesk::hw::Reaction _reaction_1, ::cpptesk::hw::Reaction _reaction_2)
    // Compare the reactions
    {
        int _match = FALSE_MATCH;
        
        if((_reaction_1.getError() == INCORRECT_REACTION)&&(_reaction_2.getError() == INCORRECT_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *_message1Spec = _reaction_1.getSpecMessage();
            const ::cpptesk::hw::Message *_message1Impl = _reaction_1.getImplMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *_message2Spec = _reaction_2.getSpecMessage();
            const ::cpptesk::hw::Message *_message2Impl = _reaction_2.getImplMessage();
    
            if((_message1Spec->size() == _message2Impl->size())&&(_message2Spec->size() == _message1Impl->size()))
            {
                _match = EXACT_MATCH;

                for(::cpptesk::hw::Message::const_iterator i = _message1Spec->begin(), j = _message2Impl->begin(),
                    ii = _message2Spec->begin(), jj = _message1Impl->begin();
                    i != _message1Spec->end(); i++, j++, ii++, jj++)
                {
                    const Field<> &field_Spec = (*_message1Spec).**i;
                    const Field<> &field_Impl = (*_message2Impl).**j;
                    const Field<> &field_Spec2 = (*_message2Spec).**ii;
                    const Field<> &field_Impl2 = (*_message1Impl).**jj;

                    if(!((field_Spec.equals(field_Impl))&&(field_Spec2.equals(field_Impl2))))
                    {
                        _match = FUZZY_MATCH;
                    }
                    
                }
            }
        }
        
        return _match;
    }

}}}
