/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/debug.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>

#include <algorithm>
#include <ostream>
#include <vector>

#define __CPPTESK_CHECK_ARBITER() \
   do { \
        if(!_arbiter) { \
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, "Reaction arbiter is not set for the interface " << *this); \
        } \
   } while(false)

namespace cpptesk {
namespace hw {

void Interface::request(const Process &process, const Message &message)
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:REQUEST](iface=" << *this << ")");

    Component::request(process);
    _message = (Message*)&message;

    // Increment number of stimuli
    incrementEventNumber();
}

void Interface::capture()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:CAPTURE](iface=" << *this << ")");

    Component::capture();
 }

void Interface::release()
{
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[IFACE:RELEASE](iface=" << *this << ")");

    Component::release();
    _message = 0;
}

void Interface::appendRequest(const Process &process, const Message &message)
{
    if(_arbiter) { _arbiter->appendRequest(process, message); }
    else         { _default_arbiter->appendRequest(process, message); }
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:APPEND](iface=" << *this << "): " << process);

    _request_count++;
}

bool Interface::chooseRequest(const Process &process, const Message &message)
{
    bool choose = _arbiter ? _arbiter->chooseRequest(process, message) : _default_arbiter->chooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:CHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

bool Interface::preChooseRequest(const Process &process, const Message &message)
{
    bool choose = _arbiter ? _arbiter->preChooseRequest(process, message) : _default_arbiter->preChooseRequest(process, message);
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:PRECHOOSE](iface=" << *this << "): " << process << ": " << (choose ? "YES" : "NO"));

    return choose;
}

void Interface::removeRequest(const Process &process, const Message &message)
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to remove (probably, WAIT_REACTION() is missing)");

    if(_arbiter) { _arbiter->removeRequest(process, message); }
    else         { _default_arbiter->removeRequest(process, message); }
    CPPTESK_DEBUG_PRINT(DEBUG_MORE, "[ARBITER:REMOVE](iface=" << *this << "): " << process);

    _request_count--;
    _taken = false;
}

bool Interface::takeReaction()
{
    CPPTESK_ASSERTION(_request_count > 0, "There are no requests to take");

    // Increment number of reactions
    incrementEventNumber();

    // If a request has been taken, returns false
    if(_taken) { return false; }
    // Otherwise returns true
    return _taken = true;
}

bool Interface::isModelBased() const
{
    bool isModel = _arbiter ? _arbiter->isModelBased() : _default_arbiter->isModelBased();

    return isModel;
}

}} // namespace cpptesk::hw

