/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a hardware interface
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/arbiter/arbiter.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>

#include <algorithm>
#include <ostream>
#include <string>
#include <vector>

///////////////////////////////////////////////////////////////////////////////
// Capturing/Releasing Interfaces
///////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CAPTURE_IFACE
 * Captures an interface.
 */
#define CPPTESK_CAPTURE_IFACE() \
    do { \
        CPPTESK_CAPTURE_COMPONENT(CPPTESK_GET_IFACE()); \
    } while(false)

/** @def CPPTESK_RELEASE_IFACE
 * Releases an interface.
 */
#define CPPTESK_RELEASE_IFACE() \
     do { \
         CPPTESK_RELEASE_COMPONENT(CPPTESK_GET_IFACE()); \
     } while(false)

#ifdef HW_SHORT_NAMES
    #define CAPTURE_IFACE() CPPTESK_CAPTURE_IFACE()
    #define RELEASE_IFACE() CPPTESK_RELEASE_IFACE()
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

class Interface: public Component
{
public:
    enum interface_type_t
    {
        /// Input interface.
        INPUT  = 0x01,
        /// Output interface.
        OUTPUT = 0x02,
        /// Input/output interface (in particular, channel).
        INOUT  = INPUT | OUTPUT
    };

    Interface(interface_type_t type = INOUT, const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME):
        Component(name), _message(0), _type(type), _arbiter(0), _default_arbiter(Arbiter::createDefaultArbiter()),
        _request_count(0), _taken(false), _target(0), _event_count(0), _failure_ignore_types(0), _reaction_timeout(0),
        _unexpected_reaction_timeout(0), _reaction_timeout_is_valid(false), _unexpected_reaction_timeout_is_valid(false) {}

    virtual ~Interface() { delete _default_arbiter; }

    void registerArbiter(Arbiter &arbiter) { _arbiter = &arbiter; }

    bool isInput()  const { return (_type & INPUT)  != 0; }
    bool isOutput() const { return (_type & OUTPUT) != 0; }

    ///////////////////////////////////////////////////////////////////////////
    // Component interface (basically, for input interfaces)
    ///////////////////////////////////////////////////////////////////////////
    virtual void request(const Process &process, const Message &message);
    virtual void capture();
    virtual void release();

    const Message* getMessage() const { return _message; }

    ///////////////////////////////////////////////////////////////////////////
    // Arbitration interface (basically, for output interfaces)
    ///////////////////////////////////////////////////////////////////////////
    void appendRequest    (const Process &process, const Message &message);
    bool chooseRequest    (const Process &process, const Message &message);
    bool preChooseRequest (const Process &process, const Message &message);
    void removeRequest    (const Process &process, const Message &message);

    Arbiter* getArbiter() { return _arbiter ? _arbiter : _default_arbiter; }

    unsigned requestNumber() const { return _request_count; }

    /// Returns false if a reaction has been taken by someone else.
    bool takeReaction();
    bool isReactionTaken() const { return _taken; }

    void setTarget(const Interface &target) { _target = &target; }
    const Interface* getTarget() const { return _target; }

    /// Returns false if the connected arbiter is adaptive.
    bool isModelBased() const;

    ///////////////////////////////////////////////////////////////////////////
    // Statistics on events (updated in takeReaction() and request())
    ///////////////////////////////////////////////////////////////////////////
    void incrementEventNumber() { _event_count++; }
    unsigned eventNumber() const { return _event_count; }

    ///////////////////////////////////////////////////////////////////////////
    // Ignorance of selected failure types
    ///////////////////////////////////////////////////////////////////////////
    void setFailureIgnoreTypes(unsigned mask) { _failure_ignore_types = mask; }
    unsigned getFailureIgnoreTypes() { return _failure_ignore_types; }

    ///////////////////////////////////////////////////////////////////////////
    // Reaction timeouts of particulair interface
    ///////////////////////////////////////////////////////////////////////////
    void setReactionTimeout(time_t time) { _reaction_timeout = time; _reaction_timeout_is_valid = true; }
    time_t getReactionTimeout() { return _reaction_timeout; }
    bool isReactionTimeoutValid() { return _reaction_timeout_is_valid; }

    void setUnexpectedReactionTimeout(time_t time) { _unexpected_reaction_timeout = time; _unexpected_reaction_timeout_is_valid = true; }
    time_t getUnexpectedReactionTimeout() { return _unexpected_reaction_timeout; }
    bool isUnexpectedReactionTimeoutValid() { return _unexpected_reaction_timeout_is_valid; }

protected:
    /// Pointer to a message being transmitted.
    const Message *_message;
    /// Interface type (input, output or input/output).
    interface_type_t _type;
    /// Reaction arbiter associated with the interface (for output interfaces only).
    Arbiter *_arbiter;
    /// Default adaptive arbiter.
    Arbiter *_default_arbiter;
    /// Number of requests competing for a reaction (for output interfaces only).
    unsigned _request_count;
    /// Flag that shows if a reaction is taken or not (for output interfaces only).
    bool _taken;
    /// Target interface (for channels, especially broadcast ones).
    const Interface *_target;
    /// Number of stimuli or reactions.
    unsigned _event_count;
    /// Ignoring failure mask.
    unsigned _failure_ignore_types;
    /// Missing reaction timeout.
    time_t _reaction_timeout;
    /// Unexpected reaction timeout.
    time_t _unexpected_reaction_timeout;
    /// Missing reaction timeout has been assigned once.
    bool _reaction_timeout_is_valid;
    /// Unexpected reaction timeout has been assigned once.
    bool _unexpected_reaction_timeout_is_valid;
};

class InputInterface: public Interface
{
public:
    InputInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(INPUT, name) {}
    virtual ~InputInterface() {}
};

class OutputInterface: public Interface
{
public:
    OutputInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(OUTPUT, name) {}
    virtual ~OutputInterface() {}
};

class InoutInterface: public Interface
{
public:
    InoutInterface(const std::string &name = __CPPTESK_DEFAULT_COMPONENT_NAME): Interface(INOUT, name) {}
    virtual ~InoutInterface() {}
};

inline std::ostream& operator << (std::ostream &out, const Interface &iface)
{
    return out << iface.getName();
}

////////////////////////////////////////////////////////////////////////////////
// Functions for counting stimuli and reactions
////////////////////////////////////////////////////////////////////////////////

inline unsigned stimulus_number(const InputInterface &iface)
{
    return iface.eventNumber();
}

template<unsigned array_size>
inline unsigned stimulus_number(InputInterface iface[array_size])
{
    unsigned count = 0;

    for(size_t i = 0; i < array_size; i++)
        { count += iface[i].eventNumber(); }

    return count;
}

template<unsigned i_size, unsigned j_size>
inline unsigned stimulus_number(const InputInterface iface[i_size][j_size])
{
    unsigned count = 0;

    for(size_t i = 0; i < i_size; i++)
    for(size_t j = 0; j < j_size; j++)
        { count += iface[i][j].eventNumber(); }

    return count;
}

inline unsigned reaction_number(const OutputInterface &iface)
{
    return iface.eventNumber();
}

template<unsigned array_size>
inline unsigned reaction_number(OutputInterface iface[array_size])
{
    unsigned count = 0;

    for(size_t i = 0; i < array_size; i++)
        { count += iface[i].eventNumber(); }

    return count;
}

template<unsigned i_size, unsigned j_size>
inline unsigned reaction_number(const OutputInterface iface[i_size][j_size])
{
    unsigned count = 0;

    for(size_t i = 0; i < i_size; i++)
    for(size_t j = 0; j < j_size; j++)
        { count += iface[i][j].eventNumber(); }

    return count;
}

}} // namespace cpptesk::hw
