/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Model of a message
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <hw/debug.hpp>
#include <hw/message.hpp>

namespace cpptesk {
namespace hw {

EmptyMessage* EmptyMessage::_instance = NULL;
const std::string Message::COMPARE_OK;

void Message::copyFields(const Message &rhs)
{
    for(size_t i = 0; i < _fields.size(); i++)
    {
        CPPTESK_ASSERTION(i < rhs._fields.size(), "Different number of fields");

        field_t lhs_field = _fields[i];
        field_t rhs_field = rhs._fields[i];

        //this->*lhs_field = rhs.*rhs_field;
        (this->*lhs_field)._copy(rhs.*rhs_field);
    }
}

Message::Message(const Message &rhs):
    Randomizable(),
    Cloneable(),
    _name(rhs._name),
    // Copy references to the subclass's fields
    _fields(rhs._fields),
    _valid(rhs._valid),
    _explanation(rhs._explanation),
    _sysdata(rhs._sysdata),
    _compare_flags(rhs._compare_flags),
    _detected(rhs._detected),
    _unexpected(rhs._unexpected)
{
    // IMPORTANT: The fields by themselves are not copied: it is done automatically
    // by a subclass's default copy constructor. If subclass copy constructor is
    // overloaded, it should invoke void Message::copyFields(const Message &rhs).
}

Message& Message::operator = (const Message &rhs)
{
    if(&rhs != this)
    {
        _name = rhs._name;
        _valid = rhs._valid;
        _explanation = rhs._explanation;
        _sysdata = rhs._sysdata;
        _compare_flags = rhs._compare_flags;
        _detected = rhs._detected;
        _unexpected = rhs._unexpected;

        copyFields(rhs);
    }

    return *this;
}

void Message::setValidFields(bool valid)
{
    for(size_t i = 0; i < _fields.size(); i++)
        { (this->*_fields[i]).setValid(valid); }
}

void Message::validateFields()
{
    setValidFields(true);
}

void Message::invalidateFields()
{
    setValidFields(false);
}

/// Compares a message with another one and returns non-empty string if they are not equal.
std::string Message::compare(const Message &message) const
{
    std::stringstream buffer;
    bool first_mismatch = true;

    CPPTESK_ASSERTION(_fields.size() == message._fields.size(), "Different number of fields");

    if(_fields.size() == 0)
    {
        CPPTESK_DEBUG_PRINT(DEBUG_WARN, toString() << "Message " << toString() << " has no fieds to compare");
        return COMPARE_OK;
    }

    // If this message (which is a reference one) is invalid,
    // then comparison is not carried out (i.e. invalid = unknown)
    if(!(isValid() && message.isValid()))
    {
        CPPTESK_DEBUG_PRINT(DEBUG_WARN, (!isValid() ? "Specification message " : "Implementation message ") << toString() << " is invalid");
        return COMPARE_OK;
    }

    for(size_t i = 0; i < _fields.size(); i++)
    {
        field_t lhs_field = _fields[i];
        field_t rhs_field = message._fields[i];

        CPPTESK_ASSERTION((this->*lhs_field).isComparable() == (message.*rhs_field).isComparable(),
            "spec_field.isComparable() != impl_field.isComparable()");

        if((this->*lhs_field).isComparable())
	{
            std::string error = (this->*lhs_field).compare(message.*rhs_field);

            if(!error.empty())
            {
                if((_compare_flags & FIRST_MISMATCH_ONLY) != 0)
                    { return error; }

                buffer << (first_mismatch ? "" : ", ") << error;
                first_mismatch = false;
            }
	}
    }

    return buffer.str();
}

std::string Message::forceCompare(const Message &message) const
{
    std::stringstream buffer;
    bool first_mismatch = true;

    for(size_t i = 0; i < _fields.size(); i++)
    {
        CPPTESK_ASSERTION(i < message._fields.size(), "Different number of fields");

        field_t lhs_field = _fields[i];
        field_t rhs_field = message._fields[i];

        if((this->*lhs_field).isComparable())
	{
	    // IMPORTANT: forceCompare() is used instead of compare()
            std::string error = (this->*lhs_field).forceCompare(message.*rhs_field);

            if(!error.empty())
            {
                if((_compare_flags & FIRST_MISMATCH_ONLY) != 0)
                    { return error; }

                buffer << (first_mismatch ? "" : ", ") << error;
                first_mismatch = false;
            }
	}
    }

    return buffer.str();
}

/// Returns a string representation of a message.
std::string Message::toString() const
{
    std::stringstream buffer;

    buffer << getName() << "{";
    {
        if(isValid())
        {
            for(size_t i = 0; i < _fields.size(); i++)
            {
                buffer << (this->*_fields[i]).toString();

                if(i < _fields.size() - 1)
	            { buffer << " "; }
            }
        }
	else
        {
            buffer << "<unknown>";
        }
    }
    buffer << "}";

    return buffer.str();
}

/// Returns a string representation of a message.
std::string Message::forceToString() const
{
    std::stringstream buffer;

    buffer << getName() << "{";
    {
        if(!isValid())
            { buffer << "<invalid>:"; }

        for(size_t i = 0; i < _fields.size(); i++)
        {
            buffer << (this->*_fields[i]).forceToString();

            if(i < _fields.size() - 1)
                { buffer << " "; }
        }
    }
    buffer << "}";

    return buffer.str();
}

/// Randomizes a message data.
void Message::randomize()
{
    for(fields_list_t::iterator i = _fields.begin(); i != _fields.end(); i++)
    {
        if((this->**i).isRandomizable())
            { (this->**i).randomize(); }
    }
}

}} // namespace cpptesk::hw

