/*
 * Copyright 2010-2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Basic model of a synchronous hardware design
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/component.hpp>
#include <hw/debug.hpp>
#include <hw/history.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>
#include <hw/params.hpp>
#include <hw/process.hpp>
#include <hw/random.hpp>
#include <hw/signals.hpp>
#include <hw/stack.hpp>
#include <hw/synchronizer.hpp>
#include <hw/time.hpp>
#include <hw/test_tracer.hpp>
#include <hw/unused.hpp>
#include <hw/arbiter/arbiter.hpp>
#include <hw/sequence/manager.hpp>

#include <tracer/tracer.hpp>
#include <ts/coverage.hpp>

#include <algorithm>
#include <map>
#include <iostream>
#include <sstream>
#include <list>
#include <vector>

#include <assert.h>
#include <stdint.h>

#if defined(WIN32)
    #include <hw/utils/win32/ucontext.h>
#elif defined(__GNUC__)
    #include <ucontext.h>
#else
    #error "unknown platform"
#endif

extern "C" {
    extern void vsim_putUInt64Value(const char *reg, uint64_t val);
    extern uint64_t vsim_getUInt64Value(const char *reg);
    extern void vsim_putUInt8Value(const char *reg, uint8_t val);
    extern uint8_t vsim_getUInt8Value(const char *reg);
}

////////////////////////////////////////////////////////////////////////////////
// Return Types of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_PROCESS
 * Keywork marking a model process.
 */
#define CPPTESK_PROCESS void

/** @def CPPTESK_STIMULUS
 * Keywork marking a stimulus.
 */
#define CPPTESK_STIMULUS CPPTESK_PROCESS

/** @def CPPTESK_REACTION
 * Keywork marking a reaction.
 */
#define CPPTESK_REACTION CPPTESK_PROCESS

#ifdef HW_SHORT_NAMES
    #define PROCESS  CPPTESK_PROCESS
    #define STIMULUS CPPTESK_STIMULUS
    #define REACTION CPPTESK_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Model Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MODEL
 * Declares a model of a hardware component.
 */
#define CPPTESK_MODEL(UserModel) \
    struct UserModel##Name { static const char* name() { return #UserModel; } }; \
    struct UserModel: public ::cpptesk::hw::Model<UserModel, UserModel##Name>

#define CPPTESK_SIGNAL_MODEL(UserModel, Inputs, Outputs) \
    CPPTESK_MODEL(UserModel), public ::cpptesk::hw::IOSignals<Inputs, Outputs>

#ifdef HW_SHORT_NAMES
    #define MODEL        CPPTESK_MODEL
    #define SIGNAL_MODEL CPPTESK_SIGNAL_MODEL
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PROCESS
 * Returns a reference to a process parameter.
 */
#define CPPTESK_GET_PROCESS() process

/** @def CPPTESK_GET_IFACE
 * Returns a reference to an interface parameter.
 */
#define CPPTESK_GET_IFACE() iface

/** @def CPPTESK_GET_MESSAGE
 * Returns a reference to a message parameter.
 */
#define CPPTESK_GET_MESSAGE() message

/** @def CPPTESK_CAST_MESSAGE
 * Casts a message parameter to a given type and returns a reference.
 */
#define CPPTESK_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

#define CPPTESK_CONST_CAST_MESSAGE(MessageType) \
    ::cpptesk::hw::_cpptesk_const_cast_message<MessageType>(CPPTESK_GET_MESSAGE(), __FILE__, __LINE__)

/** @def CPPTESK_GET_PARAMETERS
 * Returns a comma-separated list of parameters.
 */
#define CPPTESK_GET_PARAMETERS() \
    CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE()

#ifdef HW_SHORT_NAMES
    #define GET_PROCESS()                   CPPTESK_GET_PROCESS()
    #define GET_IFACE()                     CPPTESK_GET_IFACE()
    #define GET_MESSAGE()                   CPPTESK_GET_MESSAGE()
    #define CAST_MESSAGE(MessageType)       CPPTESK_CAST_MESSAGE(MessageType)
    #define CONST_CAST_MESSAGE(MessageType) CPPTESK_CONST_CAST_MESSAGE(MessageType)
    #define GET_PARAMETERS()                CPPTESK_GET_PARAMETERS()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DECLARE_PROCESS
 * Declares a process within a model class.
 */
#define CPPTESK_DECLARE_PROCESS(process_name) \
    virtual CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
 * Declares a non-virtal process within a model class.
 */
#define CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS
 * Declares a stimulus within a model class.
 */
#define CPPTESK_DECLARE_STIMULUS(stimulus_name) \
    virtual CPPTESK_STIMULUS stimulus_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_PRECONDITION
 * Declares a stimulus precondition within a model class.
 */
#define CPPTESK_DECLARE_PRECONDITION(stimulus_name) \
    virtual bool pre_##stimulus_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DECLARE_STIMULUS_WITH_PRE \
 * Declares a stimulus and its precondition.
 */
#define CPPTESK_DECLARE_STIMULUS_WITH_PRE(stimulus_name) \
    CPPTESK_DECLARE_PRECONDITION(stimulus_name); \
    CPPTESK_DECLARE_STIMULUS(stimulus_name)

/** @def CPPTESK_DECLARE_REACTION
 * Declares a reaction within a model class.
 */
#define CPPTESK_DECLARE_REACTION(reaction_name) \
    virtual CPPTESK_REACTION reaction_name(::cpptesk::hw::Process &CPPTESK_GET_PROCESS(), \
        ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

#define CPPTESK_DECLARE_INPUT_ADAPTER(adapter_name)  CPPTESK_DECLARE_PROCESS(adapter_name)
#define CPPTESK_DEFINE_INPUT_ADAPTER(adapter_name)   CPPTESK_DEFINE_PROCESS(adapter_name)

#define CPPTESK_DECLARE_OUTPUT_ADAPTER(adapter_name) CPPTESK_DECLARE_PROCESS(adapter_name)
#define CPPTESK_DEFINE_OUTPUT_ADAPTER(adapter_name)  CPPTESK_DEFINE_PROCESS(adapter_name)

#ifdef HW_SHORT_NAMES
    #define DECLARE_PROCESS             CPPTESK_DECLARE_PROCESS
    #define DECLARE_NON_VIRTUAL_PROCESS CPPTESK_DECLARE_NON_VIRTUAL_PROCESS
    #define DECLARE_STIMULUS            CPPTESK_DECLARE_STIMULUS
    #define DECLARE_PRECONDITION        CPPTESK_DECLARE_PRECONDITION
    #define DECLARE_STIMULUS_WITH_PRE   CPPTESK_DECLARE_STIMULUS_WITH_PRE
    #define DECLARE_REACTION            CPPTESK_DECLARE_REACTION
    #define DECLARE_INPUT_ADAPTER       CPPTESK_DECLARE_INPUT_ADAPTER
    #define DEFINE_INPUT_ADAPTER        CPPTESK_DEFINE_INPUT_ADAPTER
    #define DECLARE_OUTPUT_ADAPTER      CPPTESK_DECLARE_OUTPUT_ADAPTER
    #define DEFINE_OUTPUT_ADAPTER       CPPTESK_DEFINE_OUTPUT_ADAPTER
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Declaration of Special-Purpose Fields
////////////////////////////////////////////////////////////////////////////////

#if CPPTESKHW_CHECK_VERSION(111010)

// IMPORTANT: 'iface' should start the default adapter name
#define CPPTESK_DEFAULT_INPUT_ADAPTER(iface)  iface##_serialize
#define CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface) iface##_deserialize

#define CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) MessageType_##iface
// IMPORTANT: 'iface' should end the message name to support interface arrays.
#define CPPTESK_GET_IFACE_MESSAGE(iface)      message_##iface

#define CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) \
    /* Check presence of the iface */ \
    typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_input_##iface; \
    CPPTESK_DECLARE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(iface))
#define CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) \
    /* Check presence of the iface */ \
    typedef CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) __default_output_##iface; \
    CPPTESK_DECLARE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface))

#define CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER(AdapterClass_iface) \
    CPPTESK_DEFINE_INPUT_ADAPTER(CPPTESK_DEFAULT_INPUT_ADAPTER(AdapterClass_iface))
#define CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface) \
    CPPTESK_DEFINE_OUTPUT_ADAPTER(CPPTESK_DEFAULT_OUTPUT_ADAPTER(AdapterClass_iface))

#define __CPPTESK_DECLARE_IFACE(iface_class, iface, msg_type) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface); \
    iface_class iface

#define __CPPTESK_DECLARE_IFACE_ARRAY(iface_class, iface, msg_type, array_size) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[array_size]; \
    iface_class iface[array_size]

#define __CPPTESK_DECLARE_IFACE_ARRAY_2D(iface_class, iface, msg_type, i_size, j_size) \
    typedef msg_type CPPTESK_GET_IFACE_MESSAGE_TYPE(iface); \
    CPPTESK_GET_IFACE_MESSAGE_TYPE(iface) CPPTESK_GET_IFACE_MESSAGE(iface)[i_size][j_size]; \
    iface_class iface[i_size][j_size]

#define CPPTESK_DECLARE_IFACE(iface, msg_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::Interface, iface, msg_type)
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, msg_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::Interface, iface, msg_type, array_size)
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, msg_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::Interface, iface, msg_type, i_size, j_size)

#define CPPTESK_DECLARE_INPUT(iface, msg_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InputInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, msg_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InputInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, msg_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InputInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); }

#define CPPTESK_DECLARE_OUTPUT(iface, msg_type, arbiter_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::OutputInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::OutputInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::OutputInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

#define CPPTESK_DECLARE_INOUT(iface, msg_type, arbiter_type) \
    __CPPTESK_DECLARE_IFACE(::cpptesk::hw::InoutInterface, iface, msg_type); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER(arbiter_type, iface##_matcher);
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, msg_type, arbiter_type, array_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY(::cpptesk::hw::InoutInterface, iface, msg_type, array_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY(arbiter_type, iface##_matcher, array_size);
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, msg_type, arbiter_type, i_size, j_size) \
    __CPPTESK_DECLARE_IFACE_ARRAY_2D(::cpptesk::hw::InoutInterface, iface, msg_type, i_size, j_size); \
    CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER(iface)  { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER(iface) { CPPTESK_START_PROCESS(); CPPTESK_STOP_PROCESS(); } \
    CPPTESK_DECLARE_ARBITER_ARRAY_2D(arbiter_type, iface##_matcher, i_size, j_size);

#else

/** @def CPPTESK_DECLARE_IFACE
 * Declares a model interface.
 */
#define CPPTESK_DECLARE_IFACE(iface) \
    ::cpptesk::hw::Interface iface
#define CPPTESK_DECLARE_IFACE_ARRAY(iface, array_size) \
    ::cpptesk::hw::Interface iface[array_size]
#define CPPTESK_DECLARE_IFACE_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::Interface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_INPUT
 * Declares an input interface.
 */
#define CPPTESK_DECLARE_INPUT(iface) \
    ::cpptesk::hw::InputInterface iface
#define CPPTESK_DECLARE_INPUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::InputInterface iface[array_size]
#define CPPTESK_DECLARE_INPUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::InputInterface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_OUTPUT
 * Declares an output interface.
 */
#define CPPTESK_DECLARE_OUTPUT(iface) \
    ::cpptesk::hw::OutputInterface iface
#define CPPTESK_DECLARE_OUTPUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::OutputInterface iface[array_size]
#define CPPTESK_DECLARE_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::OutputInterface iface[i_size][j_size]

/** @def CPPTESK_DECLARE_INOUT
 * Declares an input/output interface.
 */
#define CPPTESK_DECLARE_INOUT(iface) \
    ::cpptesk::hw::InoutInterface iface
#define CPPTESK_DECLARE_INOUT_ARRAY(iface, array_size) \
    ::cpptesk::hw::InoutInterface iface[array_size]
#define CPPTESK_DECLARE_INOUT_ARRAY_2D(iface, i_size, j_size) \
    ::cpptesk::hw::InoutInterface iface[i_size][j_size]

#endif // CPPTESKHW_CHECK_VERSION(111010)

/** @def CPPTESK_DECLARE_ARBITER
 * Declares a primary reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_ARBITER(type, arbiter) \
    type arbiter

#define CPPTESK_DECLARE_ARBITER_ARRAY(type, arbiter, array_size) \
    type arbiter[array_size]

#define CPPTESK_DECLARE_ARBITER_ARRAY_2D(type, arbiter, i_size, j_size) \
    type arbiter[i_size][j_size]

/** @def CPPTESK_DECLARE_TWO_LEVEL_ARBITER
 * Declares two-level reaction arbiter (matcher).
 */
#define CPPTESK_DECLARE_TWO_LEVEL_ARBITER(A1, A2, arbiter) \
    ::cpptesk::hw::TwoLevelArbiter<A1, A2> arbiter

#ifdef HW_SHORT_NAMES
    #define GET_IFACE_MESSAGE_TYPE         CPPTESK_GET_IFACE_MESSAGE_TYPE
    #define GET_IFACE_MESSAGE              CPPTESK_GET_IFACE_MESSAGE
    #define DEFAULT_INPUT_ADAPTER          CPPTESK_DEFAULT_INPUT_ADAPTER
    #define DEFAULT_OUTPUT_ADAPTER         CPPTESK_DEFAULT_OUTPUT_ADAPTER
    #define DECLARE_DEFAULT_INPUT_ADAPTER  CPPTESK_DECLARE_DEFAULT_INPUT_ADAPTER
    #define DECLARE_DEFAULT_OUTPUT_ADAPTER CPPTESK_DECLARE_DEFAULT_OUTPUT_ADAPTER
    #define DEFINE_DEFAULT_INPUT_ADAPTER   CPPTESK_DEFINE_DEFAULT_INPUT_ADAPTER
    #define DEFINE_DEFAULT_OUTPUT_ADAPTER  CPPTESK_DEFINE_DEFAULT_OUTPUT_ADAPTER
    #define DECLARE_IFACE                  CPPTESK_DECLARE_IFACE
    #define DECLARE_IFACE_ARRAY            CPPTESK_DECLARE_IFACE_ARRAY
    #define DECLARE_IFACE_ARRAY_2D         CPPTESK_DECLARE_IFACE_ARRAY_2D
    #define DECLARE_INPUT                  CPPTESK_DECLARE_INPUT
    #define DECLARE_INPUT_ARRAY            CPPTESK_DECLARE_INPUT_ARRAY
    #define DECLARE_INPUT_ARRAY_2D         CPPTESK_DECLARE_INPUT_ARRAY_2D
    #define DECLARE_OUTPUT                 CPPTESK_DECLARE_OUTPUT
    #define DECLARE_OUTPUT_ARRAY           CPPTESK_DECLARE_OUTPUT_ARRAY
    #define DECLARE_OUTPUT_ARRAY_2D        CPPTESK_DECLARE_OUTPUT_ARRAY_2D
    #define DECLARE_INOUT                  CPPTESK_DECLARE_INOUT
    #define DECLARE_INOUT_ARRAY            CPPTESK_DECLARE_INOUT_ARRAY
    #define DECLARE_INOUT_ARRAY_2D         CPPTESK_DECLARE_INOUT_ARRAY_2D
    #define DECLARE_ARBITER                CPPTESK_DECLARE_ARBITER
    #define DECLARE_ARBITER_ARRAY          CPPTESK_DECLARE_ARBITER_ARRAY
    #define DECLARE_ARBITER_ARRAY_2D       CPPTESK_DECLARE_ARBITER_ARRAY_2D
    #define DECLARE_TWO_LEVEL_ARBITER      CPPTESK_DECLARE_TWO_LEVEL_ARBITER
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Definition of Special-Purpose Methods
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_PROCESS_SIGNATURE() \
    ::cpptesk::hw::Process   &CPPTESK_GET_PROCESS(), \
    ::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
    ::cpptesk::hw::Message   &CPPTESK_GET_MESSAGE()

/** @def CPPTESK_DEFINE_PROCESS
 * Defines a process.
 */
#define CPPTESK_DEFINE_PROCESS(process_name) \
    CPPTESK_PROCESS process_name(CPPTESK_PROCESS_SIGNATURE())

/** @def CPPTESK_DEFINE_STIMULUS
 * Defines a stimulus.
 */
#define CPPTESK_DEFINE_STIMULUS(stimulus_name) \
    CPPTESK_STIMULUS stimulus_name(CPPTESK_PROCESS_SIGNATURE())

/** @def CPPTESK_DEFINE_PRECONDITION
 * Defines a stimulus precondition.
 */
#define CPPTESK_DEFINE_PRECONDITION(pre_name) \
    bool pre_name(::cpptesk::hw::Interface &CPPTESK_GET_IFACE(), \
        ::cpptesk::hw::Message &CPPTESK_GET_MESSAGE())

/** @def CPPTESK_DEFINE_REACTION
 * Defines a reaction within a model class.
 */
#define CPPTESK_DEFINE_REACTION(reaction_name) \
    CPPTESK_REACTION reaction_name(CPPTESK_PROCESS_SIGNATURE())

#ifdef HW_SHORT_NAMES
    #define PROCESS_SIGNATURE   CPPTESK_PROCESS_SIGNATURE
    #define DEFINE_PROCESS      CPPTESK_DEFINE_PROCESS
    #define DEFINE_STIMULUS     CPPTESK_DEFINE_STIMULUS
    #define DEFINE_PRECONDITION CPPTESK_DEFINE_PRECONDITION
    #define DEFINE_REACTION     CPPTESK_DEFINE_REACTION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Registration of Special-Purpose Components
////////////////////////////////////////////////////////////////////////////////

// Component
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_COMPONENT
 * Registers a component in a model.
 */
#define CPPTESK_ADD_COMPONENT(component) \
    do { \
        CPPTESK_ASSERTION(!(component).getParent(), "Component " << #component << " has been already registered"); \
        (component).updateName(#component); \
        (component).setParent(this); \
        registerComponent(component); \
    } while(false)

// Interface
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_ADD_IFACE
 * Registers an interface in a model.
 */
#define CPPTESK_ADD_IFACE(iface) \
    do { \
        CPPTESK_ADD_COMPONENT(iface); \
        /* Update interface coverage */ \
        __CPPTESK_IFACE_COVERAGE::add(&iface, iface.getName()); \
    } while(false)

#define CPPTESK_ADD_IFACE_ARRAY(iface, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            std::stringstream __cpptesk_buffer; \
            __cpptesk_buffer << #iface << "[" << std::dec << __cpptesk_i << std::hex << "]"; \
            iface[__cpptesk_i].setName(__cpptesk_buffer.str()); \
            CPPTESK_ADD_IFACE(iface[__cpptesk_i]); \
        } \
    } while(false)

#define CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            std::stringstream __cpptesk_buffer; \
            __cpptesk_buffer << #iface << "[" << std::dec << __cpptesk_i << std::hex << "][" << __cpptesk_j << "]"; \
            iface[__cpptesk_i][__cpptesk_j].setName(__cpptesk_buffer.str()); \
            CPPTESK_ADD_IFACE(iface[__cpptesk_i][__cpptesk_j]); \
        } \
    } while(false)

// Interface Adapter
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_SET_IFACE_ARRAY_ADAPTER(IFACE_ADAPTER, iface, full_method_name, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i], full_method_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(IFACE_ADAPTER, iface, full_method_name, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i][__cpptesk_j], full_method_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_ADAPTER_USER_LISTENER(IFACE_ADAPTER, iface, full_method_name, full_listener_name, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i], full_method_name, full_listener_name); \
        } \
    } while(false)

#define __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER_USER_LISTENER(IFACE_ADAPTER, iface, full_method_name, full_listener_name, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_##IFACE_ADAPTER(iface[__cpptesk_i][__cpptesk_j], full_method_name, full_listener_name); \
        } \
    } while(false)

/** @def CPPTESK_SET_INPUT_ADAPTER
 * Registers an input interface adapter (aka serializer).
 */
#define CPPTESK_SET_INPUT_ADAPTER(iface, full_method_name) \
    registerInputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

#define CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(INPUT_ADAPTER, iface, full_method_name, array_size)
#define CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(INPUT_ADAPTER, iface, full_method_name, i_size, j_size)

#if CPPTESKHW_CHECK_VERSION(111010)

#define CPPTESK_SET_DEFAULT_INPUT_ADAPTER(iface) \
    CPPTESK_SET_INPUT_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_INPUT_ARRAY_ADAPTER(iface, array_size) \
    CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER(iface, i_size, j_size) \
    CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, CPPTESK_DEFAULT_INPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_PARALLEL_EX(model_t::listener, iface, CPPTESK_GET_IFACE_MESSAGE(iface), ::cpptesk::hw::EMPTY_MODEL::START_DONT_UPDATE_IFACE_COVERAGE); \
    } while(false)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER(iface) \
    CPPTESK_SET_OUTPUT_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER(iface, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER(iface, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(OUTPUT_ADAPTER_NO_LISTENER, iface, full_method_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(OUTPUT_ADAPTER_NO_LISTENER, iface, full_method_name, i_size, j_size)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER(iface) \
    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface))
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), i_size, j_size)

#define CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_LOOP_EX(full_listener_name, iface, ::cpptesk::hw::EMPTY_MODEL::START_DONT_UPDATE_IFACE_COVERAGE | ::cpptesk::hw::HIGHEST_PRIORITY); \
    } while(false)

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER_USER_LISTENER(OUTPUT_ADAPTER_USER_LISTENER, iface, full_method_name, full_listener_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER_USER_LISTENER(OUTPUT_ADAPTER_USER_LISTENER, iface, full_method_name, full_listener_name, i_size, j_size)

#define CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER(iface, full_listener_name) \
    CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), full_listener_name)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, full_listener_name, array_size) \
    CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_ADAPTER(model_t::iface), array_size)
#define CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, full_listener, i_size, j_size) \
    CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_USER_LISTENER(iface, CPPTESK_DEFAULT_OUTPUT_LISTENER(model_t::iface), full_listener_name, i_size, j_size)

#else

/** @def CPPTESK_SET_OUTPUT_ADAPTER
 * Registers an output interface adapter (aka deserializer).
 */
#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

#endif // CPPTESKHW_VERSION_CHECK

#define CPPTESK_SET_OUTPUT_ARRAY_ADAPTER(iface, full_method_name, array_size) \
    __CPPTESK_SET_IFACE_ARRAY_ADAPTER(OUTPUT_ADAPTER, iface, full_method_name, array_size)
#define CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER(iface, full_method_name, i_size, j_size) \
    __CPPTESK_SET_IFACE_ARRAY_2D_ADAPTER(OUTPUT_ADAPTER, iface, full_method_name, i_size, j_size)

/** @def CPPTESK_SET_INPUT_ADAPTER_OF
 * Registers an input interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_INPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerInputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

/** @def CPPTESK_SET_OUTPUT_ADAPTER_OF
 * Registers an output interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_OUTPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerOutputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

#if CPPTESKHW_CHECK_VERSION(111010)

// Input: No Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_INPUT(iface) \
    CPPTESK_ADD_IFACE(iface)

#define CPPTESK_ADD_INPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)

#define CPPTESK_ADD_INPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

// Input: Default Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_INPUT_WITH_DEFAULT_ADAPTER(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_INPUT_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface)); \
    } while(false)

#define CPPTESK_ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_INPUT_ARRAY_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface), array_size); \
    } while(false)

#define CPPTESK_ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER(iface, model_t::CPPTESK_DEFAULT_INPUT_ADAPTER(iface), i_size, j_size); \
    } while(false)

// Output: No Adapter
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_OUTPUT(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_ARBITER(iface, iface##_matcher); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_ARBITER_ARRAY(iface, iface##_matcher, array_size); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_ARBITER_ARRAY_2D(iface, iface##_matcher, i_size, j_size); \
    } while(false)

// Output: Default Adapter (No Listener)
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_ADD_OUTPUT_WITH_DEFAULT_ADAPTER(iface) \
    do { \
        CPPTESK_ADD_IFACE(iface); \
        CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface)); \
        CPPTESK_SET_ARBITER(iface, iface##_matcher); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER(iface, array_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY(iface, array_size); \
        CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface), array_size); \
        CPPTESK_SET_ARBITER_ARRAY(iface, iface##_matcher, array_size); \
    } while(false)

#define CPPTESK_ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER(iface, i_size, j_size) \
    do { \
        CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size); \
        CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER_NO_LISTENER(iface, model_t::CPPTESK_DEFAULT_OUTPUT_ADAPTER(iface), i_size, j_size); \
        CPPTESK_SET_ARBITER_ARRAY_2D(iface, iface##_matcher, i_size, j_size); \
    } while(false)

#else

#define CPPTESK_ADD_INPUT(iface) \
    CPPTESK_ADD_IFACE(iface)
#define CPPTESK_ADD_INPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)
#define CPPTESK_ADD_INPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

#define CPPTESK_ADD_OUTPUT(iface) \
    CPPTESK_ADD_IFACE(iface)
#define CPPTESK_ADD_OUTPUT_ARRAY(iface, array_size) \
    CPPTESK_ADD_IFACE_ARRAY(iface, array_size)
#define CPPTESK_ADD_OUTPUT_ARRAY_2D(iface, i_size, j_size) \
    CPPTESK_ADD_IFACE_ARRAY_2D(iface, i_size, j_size)

#endif // CPPTESKHW_CHECK_VERSION

/** @def CPPTESK_SET_ARBITER
 * Registers a primary arbiter for a given output interface.
 */
#define CPPTESK_SET_ARBITER(iface, arbiter) \
    iface.registerArbiter(arbiter)

#define CPPTESK_SET_ARBITER_ARRAY(iface, arbiter, array_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < array_size; __cpptesk_i++) \
        { \
            CPPTESK_SET_ARBITER(iface[__cpptesk_i], arbiter[__cpptesk_i]); \
        } \
    } while(false)

#define CPPTESK_SET_ARBITER_ARRAY_2D(iface, arbiter, i_size, j_size) \
    do { \
        for(int __cpptesk_i = 0; __cpptesk_i < i_size; __cpptesk_i++) \
        for(int __cpptesk_j = 0; __cpptesk_j < j_size; __cpptesk_j++) \
        { \
            CPPTESK_SET_ARBITER(iface[__cpptesk_i][__cpptesk_j], arbiter[__cpptesk_i][__cpptesk_j]); \
        } \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define ADD_COMPONENT                                     CPPTESK_ADD_COMPONENT
    #define ADD_IFACE                                         CPPTESK_ADD_IFACE
    #define ADD_IFACE_ARRAY                                   CPPTESK_ADD_IFACE_ARRAY
    #define ADD_IFACE_ARRAY_2D                                CPPTESK_ADD_IFACE_ARRAY_2D
    #define ADD_INPUT                                         CPPTESK_ADD_INPUT
    #define ADD_INPUT_ARRAY                                   CPPTESK_ADD_INPUT_ARRAY
    #define ADD_INPUT_ARRAY_2D                                CPPTESK_ADD_INPUT_ARRAY_2D
    #define ADD_OUTPUT                                        CPPTESK_ADD_OUTPUT
    #define ADD_OUTPUT_ARRAY                                  CPPTESK_ADD_OUTPUT_ARRAY
    #define ADD_OUTPUT_ARRAY_2D                               CPPTESK_ADD_OUTPUT_ARRAY_2D
    #define SET_ARBITER                                       CPPTESK_SET_ARBITER
    #define SET_ARBITER_ARRAY                                 CPPTESK_SET_ARBITER_ARRAY
    #define SET_ARBITER_ARRAY_2D                              CPPTESK_SET_ARBITER_ARRAY_2D
    #define SET_INPUT_ADAPTER                                 CPPTESK_SET_INPUT_ADAPTER
    #define SET_INPUT_ARRAY_ADAPTER                           CPPTESK_SET_INPUT_ARRAY_ADAPTER
    #define SET_INPUT_ARRAY_2D_ADAPTER                        CPPTESK_SET_INPUT_ARRAY_2D_ADAPTER
    #define SET_OUTPUT_ADAPTER                                CPPTESK_SET_OUTPUT_ADAPTER
    #define SET_OUTPUT_ARRAY_ADAPTER                          CPPTESK_SET_OUTPUT_ARRAY_ADAPTER
    #define SET_OUTPUT_ARRAY_2D_ADAPTER                       CPPTESK_SET_OUTPUT_ARRAY_2D_ADAPTER
    #define SET_DEFAULT_INPUT_ADAPTER                         CPPTESK_SET_DEFAULT_INPUT_ADAPTER
    #define SET_DEFAULT_INPUT_ARRAY_ADAPTER                   CPPTESK_SET_DEFAULT_INPUT_ARRAY_ADAPTER
    #define SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER                CPPTESK_SET_DEFAULT_INPUT_ARRAY_2D_ADAPTER
    #define SET_DEFAULT_OUTPUT_ADAPTER                        CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER                  CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER
    #define SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER               CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_2D_ADAPTER
    #define SET_INPUT_ADAPTER_OF                              CPPTESK_SET_INPUT_ADAPTER_OF
    #define SET_OUTPUT_ADAPTER_OF                             CPPTESK_SET_OUTPUT_ADAPTER_OF
#if CPPTESKHW_CHECK_VERSION(111010)
    #define ADD_INPUT_WITH_DEFAULT_ADAPTER                    CPPTESK_ADD_INPUT_WITH_DEFAULT_ADAPTER
    #define ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER              CPPTESK_ADD_INPUT_ARRAY_WITH_DEFAULT_ADAPTER
    #define ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER           CPPTESK_ADD_INPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_WITH_DEFAULT_ADAPTER                   CPPTESK_ADD_OUTPUT_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER             CPPTESK_ADD_OUTPUT_ARRAY_WITH_DEFAULT_ADAPTER
    #define ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER          CPPTESK_ADD_OUTPUT_ARRAY_2D_WITH_DEFAULT_ADAPTER
    #define SET_OUTPUT_ADAPTER_NO_LISTENER                    CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER              CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_NO_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER           CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER
    #define SET_OUTPUT_ADAPTER_USER_LISTENER                  CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER            CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_USER_LISTENER
    #define SET_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER         CPPTESK_SET_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER            CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER      CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER   CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_NO_LISTENER
    #define SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER          CPPTESK_SET_DEFAULT_OUTPUT_ADAPTER_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER    CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_USER_LISTENER
    #define SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER CPPTESK_SET_DEFAULT_OUTPUT_ARRAY_ADAPTER_2D_USER_LISTENER

#endif // CPPTESKHW_CHECK_VERSION
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Parameters
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_REACTION_TIMEOUT
 * Sets reaction timeout.
 */
#define CPPTESK_SET_REACTION_TIMEOUT(timeout) \
    setReactionTimeout(timeout)

#define CPPTESK_SET_UNEXPECTED_REACTION_TIMEOUT(timeout) \
    setUnexpectedReactionTimeout(timeout)

#ifdef HW_SHORT_NAMES
    #define SET_REACTION_TIMEOUT(timeout) CPPTESK_SET_REACTION_TIMEOUT(timeout)
    #define SET_UNEXPECTED_REACTION_TIMEOUT(timeout) CPPTESK_SET_UNEXPECTED_REACTION_TIMEOUT(timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Control Statements
////////////////////////////////////////////////////////////////////////////////

/**
 * Process status is reassigned to EXECUTING to handle nested calls.
 */
#define __CPPTESK_SYNPOINT(type, name) \
    { \
        CPPTESK_GET_PROCESS().point = ::cpptesk::hw::Process::PointDescriptor(type, name, __FILE__, __LINE__); \
        /* IMPORTANT: Timeout flag is not reset by CYCLE() and DELAY() */ \
        if(type == ::cpptesk::hw::GUARD) { CPPTESK_GET_PROCESS().timeout = false; } \
        CPPTESK_GET_PROCESS().status = ::cpptesk::hw::EXECUTING; \
    }

#define __CPPTESK_RETURN(result) \
    do { \
        /* To check presense of CPPTESK_START_PROCESS() */ \
        __CPPTESK_CHECK_START(); \
        CPPTESK_ASSERTION(CPPTESK_GET_PROCESS().call_stack_size() > 0, "There is no an opening START for the current STOP"); \
        CPPTESK_GET_PROCESS().status = (result); \
        /* External parallel process jumps to the parent process */ \
        CPPTESK_GET_PROCESS().stop_call(); \
        if(CPPTESK_GET_PROCESS().call_stack_size() == 0) { \
            suspend(&CPPTESK_GET_PROCESS()); \
        } \
        return; \
    } while(false)

////////////////////////////////////////////////////////////////////////////////
// Precondition and Postcondition
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CHECK_CONDITION
 * Returns false if a condition is not satistied. The macros is typically used
 * in precondition methods.
 */
#define CPPTESK_CHECK_CONDITION(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Condition " << #cond << " is violated"); \
        return false; \
    }

/** @def CPPTESK_PRE
 * Precondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_PRE(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Precondition " << #cond << " is violated"); \
        CPPTESK_GET_PROCESS().point.file = __FILE__; \
        CPPTESK_GET_PROCESS().point.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::UNPREDICT); \
    }

/** @def CPPTESK_POST
 * Postcondition assertion. The macros is typically used in stimulus methods.
 */
#define CPPTESK_POST(cond) \
    if(!(cond)) \
    { \
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_INFO, "Postcondition " << #cond << " is violated"); \
        CPPTESK_GET_PROCESS().point.file = __FILE__; \
        CPPTESK_GET_PROCESS().point.line = __LINE__; \
        __CPPTESK_RETURN(::cpptesk::hw::VIOLATION); \
    }

#ifdef HW_SHORT_NAMES
    #define CHECK_CONDITION(cond) CPPTESK_CHECK_CONDITION(cond)
    #define PRE(cond)             CPPTESK_PRE(cond)
    #define POST(cond)            CPPTESK_POST(cond)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Call Process
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_PREPARE_CALL(method_name) \
    prepare_call(::cpptesk::hw::Process::CallDescriptor(#method_name, __FILE__, __LINE__))

#define CPPTESK_CALL_PARALLEL(method_name, iface, message) \
    ( \
        /* IMPORTANT: Do not check existance of START() */ \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, iface, message, START_IMMEDIATELY) \
    )

#define CPPTESK_CALL_SEQUENTIAL(method_name, iface, message) \
    ( \
        __CPPTESK_CHECK_START(), \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::process_t)&model_t::method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) \
    ( \
        /* IMPORTANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, iface, message, mode) \
    )

#define CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message) \
    ( \
        /* IMPORANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::process_t)&full_method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

#define CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message) \
    ((call_mode) == PARALLEL ? \
        CPPTESK_CALL_PARALLEL(method_name, iface, message) : \
        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message))

#define CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) \
    ((call_mode) == PARALLEL ? \
        /* IMPORTANT: Parallel call does not use the process parameter */ \
        CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) : \
        /* IMPORTANT: Sequential call does not use the mode parameter */ \
        CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message, mode))

#define CPPTESK_RECV_STIMULUS(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onStimulus, iface, message)

#define CPPTESK_SEND_REACTION(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
   _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(CPPTESK_GET_IFACE()), __FILE__, __LINE__))

#define CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
    _cpptesk_set_explanation(message, time(), time() + getReactionTimeout(CPPTESK_GET_IFACE()), __FILE__, __LINE__, comment))

#define CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked */ \
        (model).start(&full_method_name, iface, message) \
    )

#define CPPTESK_CALL_PROCESS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked */ \
        (model).start(&full_method_name, iface, message, mode) \
    )

#define CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message) \
    ( \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the interface */ \
        ::cpptesk::hw::__cpptesk_assertion_file_line((iface).getStatus() == ::cpptesk::hw::Interface::IDLE, \
            "Interface " + (iface).getName() + " has been already requested", __FILE__, __LINE__), \
        /* The interface is locked */ \
        (model).start(&full_method_name, iface, message, ::cpptesk::hw::EMPTY_MODEL::START_REQUEST_IFACE) \
    )

#define CPPTESK_CALL_STIMULUS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
         /* Check the interface */ \
        ::cpptesk::hw::__cpptesk_assertion_file_line((iface).getStatus() == ::cpptesk::hw::Interface::IDLE, \
            "Interface " + (iface).getName() + " has been already requested", __FILE__, __LINE__), \
        /* The interface is locked */ \
        (model).start(&full_method_name, iface, message, ::cpptesk::hw::EMPTY_MODEL::START_REQUEST_IFACE | (mode)) \
    )

#define CPPTESK_CALL_REACTION_OF(model, full_method_name, iface, message) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked, and the process is launched IMMEDIATELY */ \
        (model).start(&full_method_name, iface, message) \
    )

#define CPPTESK_CALL_REACTION_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked, and the process is launched IMMEDIATELY */ \
        (model).start(&full_method_name, iface, message, mode) \
    )

/** @def CPPTESK_CALL_LOOP
 * Launches an infinite loop.
 */
#define CPPTESK_CALL_LOOP(method_name) \
    CPPTESK_CALL_PARALLEL(method_name, CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE)

/** @def CPPTESK_CALL_LOOP_EX
 * Launches an infinite loop (usually, a listener of an output interface).
 */
#define CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode) \
    CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE, mode)

#ifdef HW_SHORT_NAMES
    #define CALL_PARALLEL              CPPTESK_CALL_PARALLEL
    #define CALL_SEQUENTIAL            CPPTESK_CALL_SEQUENTIAL
    #define CALL_PROCESS               CPPTESK_CALL_PROCESS
    #define CALL_PARALLEL_EX           CPPTESK_CALL_PARALLEL_EX
    #define CALL_SEQUENTIAL_EX         CPPTESK_CALL_SEQUENTIAL_EX
    #define CALL_PROCESS_EX            CPPTESK_CALL_PROCESS_EX
    #define RECV_STIMULUS              CPPTESK_RECV_STIMULUS
    #define SEND_REACTION              CPPTESK_SEND_REACTION
    #define SEND_REACTION_WITH_COMMENT CPPTESK_SEND_REACTION_WITH_COMMENT
    #define CALL_PROCESS_OF            CPPTESK_CALL_PROCESS_OF
    #define CALL_PROCESS_OF_EX         CPPTESK_CALL_PROCESS_OF_EX
    #define CALL_STIMULUS_OF           CPPTESK_CALL_STIMULUS_OF
    #define CALL_STIMULUS_OF_EX        CPPTESK_CALL_STIMULUS_OF_EX
    #define CALL_REACTION_OF           CPPTESK_CALL_REACTION_OF
    #define CALL_REACTION_OF_EX        CPPTESK_CALL_REACTION_OF_EX
    #define CALL_LOOP                  CPPTESK_CALL_LOOP
    #define CALL_LOOP_EX               CPPTESK_CALL_LOOP_EX
#endif // HW_SHORT_NAMES

#define CPPTESK_CALL_STACK_OF(model) \
    ((model)._process_stack)

#define CPPTESK_CALL_STACK() \
    CPPTESK_CALL_STACK_OF(*this)

#ifdef HW_SHORT_NAMES
    #define CALL_STACK_OF(model) CPPTESK_CALL_STACK_OF(model)
    #define CALL_STACK()         CPPTESK_CALL_STACK()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Common part of model debug messages
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_COMMON_START_INFO(start_time) \
    "(" << CPPTESK_GET_IFACE() << " cycle=" << std::dec << start_time << std::hex << " pid=" << std::dec << CPPTESK_GET_PROCESS().pid << std::hex << ")"

#define __CPPTESK_COMMON_STOP_INFO(start_time, stop_time) \
    "(" << CPPTESK_GET_IFACE() << " cycle=[" << std::dec << start_time << "," << stop_time << std::hex << "] pid=" << std::dec << CPPTESK_GET_PROCESS().pid << std::hex << ")"

////////////////////////////////////////////////////////////////////////////////
// Start and Stop
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_START_LABEL $CPPTESK_START_PROCESS
#define __CPPTESK_START_VAR   _CPPTESK_START_PROCESS

#define __CPPTESK_CHECK_START() (void)__CPPTESK_START_VAR

#define __CPPTESK_INSIDE_PROCESS(process_stack) \
    /* Inform macroses that they are inside a process */ \
    bool  __cpptesk_inside_process = true; \
    void *__cpptesk_process_stack  = &(process_stack)

/** @def CPPTESK_START
 * Starts a model process.
 */
#define CPPTESK_START() \
    /* Inform macroses that they are inside a process */ \
    __CPPTESK_INSIDE_PROCESS(_process_stack); \
    \
    /* To check that CPPTESK_START() preceedes any process call */ \
    int __CPPTESK_START_VAR; \
    \
    /* Only one CPPTESK_START() is allowed within a specification function */ \
    __CPPTESK_START_LABEL: \
    \
    do { \
        /* To avoid the 'unused variable' warning */ \
        __CPPTESK_CHECK_START(); \
        /* To avoid the 'unused label' warning */ \
        if(false) { goto __CPPTESK_START_LABEL; } \
        \
        CPPTESK_GET_PROCESS().start_call(); \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:START]" : "[PARALLEL:START]") << \
            __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
        /* Check the message invariant */ \
        CPPTESK_ASSERTION(CPPTESK_GET_MESSAGE().invariant(), "Message invariant has failed: " << CPPTESK_GET_MESSAGE()); \
    } while(false)

/** @def CPPTESK_START_PROCESS
 * Starts a process.
 */
#define CPPTESK_START_PROCESS() \
    CPPTESK_START()

/** @def CPPTESK_START_STIMULUS
 * Starts a stimulus and launches the input interface adapter.
 */
#define CPPTESK_START_STIMULUS(call_mode) \
    CPPTESK_START(); \
    CPPTESK_RECV_STIMULUS(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_START_REACTION
 * Starts a reaction and launches the output interface adapter.
 */
#define CPPTESK_START_REACTION(call_mode) \
    CPPTESK_START(); \
    CPPTESK_SEND_REACTION(call_mode, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE())

/** @def CPPTESK_STOP
 * Stops a model process.
 */
#define CPPTESK_STOP() \
    do { \
        /* To check presense of CPPTESK_START() */ \
        __CPPTESK_CHECK_START(); \
         CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_MORE, (CPPTESK_GET_PROCESS().call_stack_size() > 1 ? "[SEQUENTIAL:STOP]" : "[PARALLEL:STOP]") << \
             __CPPTESK_COMMON_START_INFO(time()) << ": " << CPPTESK_GET_PROCESS()); \
        __CPPTESK_RETURN(::cpptesk::hw::COMPLETED); \
    } while(false)

/** @def CPPTESK_STOP_PROCESS
 * Stops a process.
 */
#define CPPTESK_STOP_PROCESS()  CPPTESK_STOP()

/** @def CPPTESK_STOP_STIMULUS
 * Stops a stimulus.
 */
#define CPPTESK_STOP_STIMULUS() CPPTESK_STOP()

/** @def CPPTESK_STOP_REACTION
 * Stops a reaction.
 */
#define CPPTESK_STOP_REACTION() CPPTESK_STOP()

#ifdef HW_SHORT_NAMES
    #define START()                   CPPTESK_START()
    #define START_PROCESS()           CPPTESK_START_PROCESS()
    #define START_STIMULUS(call_mode) CPPTESK_START_STIMULUS(call_mode)
    #define START_REACTION(call_mode) CPPTESK_START_REACTION(call_mode)
    #define STOP()                    CPPTESK_STOP()
    #define STOP_PROCESS()            CPPTESK_STOP_PROCESS()
    #define STOP_STIMULUS()           CPPTESK_STOP_STIMULUS()
    #define STOP_REACTION()           CPPTESK_STOP_REACTION()
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Cycle, Delay, and Guard
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY
 * Emulates a clock cycle of a unit.
 */
#define CPPTESK_CYCLE_OF(unit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        CPPTESK_ASSERTION((unit)._process_stack.size() > 1, \
            "Cannot perform CYCLE() for a unit with empty process stack"); \
        (unit).suspend(&CPPTESK_GET_PROCESS()); \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates a clock cycle.
 */
#define CPPTESK_CYCLE() \
    CPPTESK_CYCLE_OF(*this)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_DELAY_OF
 * Emulates several clock cycles of a unit.
 */
#define CPPTESK_DELAY_OF(unit, delay) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::CYCLE, "CYCLE()"); \
        while(CPPTESK_GET_PROCESS().point.timer < (::cpptesk::hw::cycle_t)(delay)) \
        { \
            CPPTESK_ASSERTION((unit)._process_stack.size() > 1, \
                "Cannot perform DELAY(" #delay ") for a unit with empty process stack"); \
            (unit).suspend(&CPPTESK_GET_PROCESS()); \
        } \
    } while(false)

/** @def CPPTESK_DELAY
 * Emulates several clock cycles.
 */
#define CPPTESK_DELAY(delay) \
    CPPTESK_DELAY_OF(*this, delay)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_OF
 * Waits unit a condition becomes true.
 */
#define CPPTESK_GUARD_OF(unit, condition) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
        { \
            CPPTESK_ASSERTION((unit)._process_stack.size() > 1, \
                "Cannot perform WAIT(" #condition ") for a unit with empty process stack"); \
            (unit).suspend(&CPPTESK_GET_PROCESS()); \
        } \
        /* Do not wait anymore */ \
        CPPTESK_GET_PROCESS().point.type = ::cpptesk::hw::CYCLE; \
        CPPTESK_GET_PROCESS().point.name = "CYCLE()"; \
    } while(false)

/** @def CPPTESK_GUARD
 * Waits until a condition becomes true.
 */
#define CPPTESK_GUARD(condition) \
    CPPTESK_GUARD_OF(*this, condition)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GUARD_TIMEOUT_OF
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit) \
    do { \
        __CPPTESK_SYNPOINT(::cpptesk::hw::GUARD, "WAIT(" #condition ")"); \
        while(!(condition)) \
        { \
            /* Timeout violation */ \
            if(CPPTESK_GET_PROCESS().point.timer >= (::cpptesk::hw::cycle_t)(limit)) \
            { \
                CPPTESK_GET_PROCESS().timeout = true; \
                onTimeoutViolation(CPPTESK_GET_PROCESS()); \
                break; \
            } \
            CPPTESK_ASSERTION((unit)._process_stack.size() > 1, \
                "Cannot perform WAIT_TIMEOUT(" #condition ", " #limit ") for a unit with empty process stack"); \
            (unit).suspend(&CPPTESK_GET_PROCESS()); \
        } \
        /* Do not wait anymore */ \
        CPPTESK_GET_PROCESS().point.type = ::cpptesk::hw::CYCLE; \
        CPPTESK_GET_PROCESS().point.name = "CYCLE()"; \
    } while(false)

/** @def CPPTESK_GUARD_TIMEOUT
 * Waits until a condition becomes true or time is over.
 */
#define CPPTESK_GUARD_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(*this, condition, limit)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_WAIT_OF
 * Alias of CPPTESK_GUARD_OF
 */
#define CPPTESK_WAIT_OF(unit, condition) \
    CPPTESK_GUARD_OF(unit, condition)

/** @def CPPTESK_WAIT
 * Alias of CPPTESK_GUARD
 */
#define CPPTESK_WAIT(condition) \
    CPPTESK_GUARD(condition)

/** @def CPPTESK_WAIT_TIMEOUT_OF
 * Alias of CPPTESK_GUARD_TIMEOUT_OF
 */
#define CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit) \
    CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)

/** @def CPPTESK_WAIT_TIMEOUT
 * Alias of CPPTESK_GUARD_TIMEOUT
 */
#define CPPTESK_WAIT_TIMEOUT(condition, limit) \
    CPPTESK_GUARD_TIMEOUT(condition, limit)

////////////////////////////////////////////////////////////////////////////////

#ifdef HW_SHORT_NAMES
    #define CYCLE_OF(unit)                           CPPTESK_CYCLE_OF(unit)
    #define CYCLE()                                  CPPTESK_CYCLE()
    #define DELAY_OF(unit, delay)                    CPPTESK_DELAY_OF(unit, delay)
    #define DELAY(delay)                             CPPTESK_DELAY(delay)
    #define GUARD_OF(unit, condition)                CPPTESK_GUARD_OF(unit, condition)
    #define GUARD(condition)                         CPPTESK_GUARD(condition)
    #define GUARD_TIMEOUT_OF(unit, condition, limit) CPPTESK_GUARD_TIMEOUT_OF(unit, condition, limit)
    #define GUARD_TIMEOUT(condition, limit)          CPPTESK_GUARD_TIMEOUT(condition, limit)
    #define WAIT_OF(unit, condition)                 CPPTESK_WAIT_OF(unit, condition)
    #define WAIT(condition)                          CPPTESK_WAIT(condition)
    #define WAIT_TIMEOUT_OF(unit, condition, limit)  CPPTESK_WAIT_TIMEOUT_OF(unit, condition, limit)
    #define WAIT_TIMEOUT(condition, limit)           CPPTESK_WAIT_TIMEOUT(condition, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Waiting Process
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_WAIT_PROCESS_STARTED(pid) \
    CPPTESK_WAIT(isFinished(pid) || isStarted(pid))

#define CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isStarted(pid), limit)

#define CPPTESK_WAIT_PROCESS_WAITING(pid) \
    CPPTESK_WAIT(isFinished(pid) || isWaiting(pid))

#define CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isWaiting(pid), limit)

#define CPPTESK_WAIT_PROCESS_ACTIVE(pid) \
    CPPTESK_WAIT(isFinished(pid) || isActive(pid))

#define CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid) || isActive(pid, limit))

#define CPPTESK_WAIT_PROCESS_FINISHED(pid) \
    CPPTESK_WAIT(isFinished(pid))

#define CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) \
    CPPTESK_WAIT_TIMEOUT(isFinished(pid), limit)

#ifdef HW_SHORT_NAMES
    #define WAIT_PROCESS_STARTED(pid)                 CPPTESK_WAIT_PROCESS_STARTED(pid)
    #define WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_STARTED_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_WAITING(pid)                 CPPTESK_WAIT_PROCESS_WAITING(pid)
    #define WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)  CPPTESK_WAIT_PROCESS_WAITING_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_ACTIVE(pid)                  CPPTESK_WAIT_PROCESS_ACTIVE(pid)
    #define WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)   CPPTESK_WAIT_PROCESS_ACTIVE_TIMEOUT(pid, limit)
    #define WAIT_PROCESS_FINISHED(pid)                CPPTESK_WAIT_PROCESS_FINISHED(pid)
    #define WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit) CPPTESK_WAIT_PROCESS_FINISHED_TIMEOUT(pid, limit)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Process Attributes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_GET_PRIORITY
 * Gets a process priority.
 */
#define CPPTESK_GET_PRIORITY() \
    ((unsigned)(CPPTESK_GET_PROCESS().getPriority()))

/** @def CPPTESK_SET_PRIORITY
 * Sets a process priority.
 */
#define CPPTESK_SET_PRIORITY(priority) \
    do { \
        CPPTESK_GET_PROCESS().setPriority((::cpptesk::hw::priority_t)priority); \
    } while(false)

/** @def CPPTESK_GET_INPUT_IFACE
 * Gets a process input interface.
 */
#define CPPTESK_GET_INPUT_IFACE() \
    (CPPTESK_GET_PROCESS().input_iface)

/** @def CPPTESK_SET_INPUT_IFACE
 * Sets a process input interface.
 */
#define CPPTESK_SET_INPUT_IFACE(iface) \
    do { \
        CPPTESK_GET_PROCESS().input_iface = &iface; \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define GET_PRIORITY()         CPPTESK_GET_PRIORITY()
    #define SET_PRIORITY(priority) CPPTESK_SET_PRIORITY(priority)
    #define GET_INPUT_IFACE()      CPPTESK_GET_INPUT_IFACE()
    #define SET_INPUT_IFACE(iface) CPPTESK_SET_INPUT_IFACE(iface)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Model Events
///////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_RAISE_EVENT
 * Raises an event.
 */
#define CPPTESK_RAISE_EVENT(id) \
    raise(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_CHECK_EVENT
 * Checks whether an event occures.
 */
#define CPPTESK_CHECK_EVENT(id) \
    check(sequence::ModelEvent(id, _model_event_manager))

/** @def CPPTESK_WAIT_EVENT
 * Waits until an event occures.
 */
#define CPPTESK_WAIT_EVENT(id) \
    CPPTESK_WAIT(CPPTESK_CHECK_EVENT(id))

/** @def CPPTESK_WAIT_EVENT_TIMEOUT
 * Waits until an event occures but not longer than an amount of time specified.
 */
#define CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout) \
    CPPTESK_WAIT_TIMEOUT(CPPTESK_CHECK_EVENT(id), timeout)

#ifdef HW_SHORT_NAMES
    #define RAISE_EVENT(id)                 CPPTESK_RAISE_EVENT(id)
    #define CHECK_EVENT(id)                 CPPTESK_CHECK_EVENT(id)
    #define WAIT_EVENT(id)                  CPPTESK_WAIT_EVENT(id)
    #define WAIT_EVENT_TIMEOUT(id, timeout) CPPTESK_WAIT_EVENT_TIMEOUT(id, timeout)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Interface Coverage
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DEFINE_IFACE_COVERAGE(cov_name) \
    struct __CPPTESK_IFACE_COVERAGE_$_enum \
    { \
        typedef cpptesk::hw::Interface *Id; \
        typedef cpptesk::ts::coverage::CoverageElement<Id> Element; \
        typedef std::vector<Element> ElementVector; \
        typedef ElementVector::iterator ElementIterator; \
        typedef cpptesk::utils::Range<ElementIterator> ElementRange; \
        \
        enum Command { ADD, GET }; \
        \
        static ElementRange main(Command cmd, const Id &id, const std::string &label) \
        { \
            static ElementVector _elements; \
            \
            if(cmd == ADD) \
                { _elements.push_back(Element(id, label)); } \
            \
            return ElementRange(_elements.begin(), _elements.end(), _elements.size()); \
        } \
        \
        static ElementRange get() { return main(GET, Id(), ""); } \
        static void add(const Id &id, const std::string &label) { main(ADD, id, label); } \
        \
        struct Matcher \
        { \
            Id id; \
            inline bool operator()(const Id& other) const { return id == other; } \
            inline Matcher& operator= (const Id& _id) { id = _id; return *this; } \
        }; \
    }; \
    struct __CPPTESK_IFACE_COVERAGE_$_name { static inline const char* name() { return cov_name; } }; \
    typedef cpptesk::ts::coverage::Coverage<__CPPTESK_IFACE_COVERAGE_$_enum, __CPPTESK_IFACE_COVERAGE_$_name> __CPPTESK_IFACE_COVERAGE;

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Basic Model Class
////////////////////////////////////////////////////////////////////////////////

template<class UserModel /* Model's sub-class */, class UserModelName> class Model: public Synchronizer
{
public:
    typedef UserModel model_t;

    const static int DEFAULT_REACTION_TIMEOUT = 10;
    const static int DEFAULT_UNEXPECTED_REACTION_TIMEOUT = 0;

    static const char* name() { return UserModelName::name();}

    typedef void (UserModel::*process_t)(Process &process, Interface &iface, Message &message);

    typedef process_t stimulus_t;
    typedef process_t reaction_t;

    enum call_mode_t
    {
        PARALLEL,
        SEQUENTIAL
    };

    enum start_mode_t
    {
        /// Calling start() adds a new process into the pipeline; the process will be executed when cycle() is called.
        START_NORMAL                     = 0x0000 | NULL_PRIORITY,
        /// Calling start() adds a new process into the pipeline and executes the first stage of the process within the current cycle.
        START_IMMEDIATELY                = 0x0100 | NULL_PRIORITY,
        /// Request interface when starting a process.
        START_REQUEST_IFACE              = 0x0200 | NULL_PRIORITY,
        /// Do not clone the message parameter.
        START_DONT_CLONE_MESSAGE         = 0x0400 | NULL_PRIORITY,
        /// Do not update the interface coverage
        START_DONT_UPDATE_IFACE_COVERAGE = 0x0800 | NULL_PRIORITY,
        /// Start very immediately
        START_THIS_CYCLE                 = 0x1000 | NULL_PRIORITY
    };

    __CPPTESK_DEFINE_IFACE_COVERAGE((std::string(model_t::name()) + " Interface Coverage").c_str());
    cpptesk::tracer::coverage::CoverageTracker coverageTracker;

    Model();
    virtual ~Model();

    void registerComponent(const Component &component);

    void registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name);

    inline void registerInputAdapter(Interface &input, process_t serializer, const std::string &name)
        { registerInterfaceAdapter(input, serializer, name); }

    inline void registerOutputAdapter(Interface &output, process_t deserializer, const std::string &name)
        { registerInterfaceAdapter(output, deserializer, name); }

    void registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name);

    inline void registerInputAdapter(Interface &input, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(input, model, adapter, name); }

    inline void registerOutputAdapter(Interface &output, model_t &model, process_t adapter, const std::string &name)
        { registerInterfaceAdapter(output, model, adapter, name); }

public:

    inline Process::CallDescriptor get_call() const { return _call; }
    inline void prepare_call(const Process::CallDescriptor &call) { _call = call; }
    inline void reset_call() { _call.reset(); }

protected:

    /**
     * Descriptor of a current process call:
     *
     * CALL_PROCESS():  prepare_call(DESCRIPTOR);
     * start():         process.prepare_call(get_call()); reset_call();
     * START_PROCESS(): process.start_call();
     * STOP_PROCESS():  process.stop_call();
     */
    Process::CallDescriptor _call;

public:

///////////////////////////////////////////////////////////////////////////////
// Process creation and destruction methods
///////////////////////////////////////////////////////////////////////////////

    /**
     * Starts a new process for a given stimulus.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, Message &message, int mode = START_NORMAL);

    /**
     * Starts a new process for a given stimulus with the empty message.
     */
    pid_t start(stimulus_t stimulus, Interface &iface, int mode = START_NORMAL);

    /**
     * Processes a stimulus within a given process. Calling start(&UserModel::stimulus, process, message)
     * is fully equivalent to userModel.stimulus(process, iface, message).
     */
    pid_t start(stimulus_t stimulus, Process &process, Interface &iface, Message &message = CPPTESK_DEFAULT_MESSAGE);

    /**
     * Stops a process.
     */
    void stop(pid_t pid);

    void reset();

    size_t size() const { return _process_table.size(); } 

///////////////////////////////////////////////////////////////////////////////
// Methods for checking process status
///////////////////////////////////////////////////////////////////////////////

    bool isStarted(pid_t pid)  const;
    bool isWaiting(pid_t pid)  const;
    bool isActive(pid_t pid)   const;
    bool isFinished(pid_t pid) const;

///////////////////////////////////////////////////////////////////////////////
// System clock modelling
///////////////////////////////////////////////////////////////////////////////

    cycle_t time() const { return _clock; }

    void cycle();
    void delay(cycle_t n);

///////////////////////////////////////////////////////////////////////////////
// History of Stimuli & Reactions
///////////////////////////////////////////////////////////////////////////////

public:

    /**
     * Register a stimulus or a reaction.
     */
    void registerStimulus(const Stimulus &stimulus);
    void registerReaction(const Reaction &reaction);

    void registerStimulus(const Interface &iface);
    void registerReaction(error_type_t error, const Interface &iface, const std::string &info, unsigned type = -1u);

    void clearReactions();

    /**
     * Returns the first error (if there is any) having occured starting from the previous call of verdict().
     * Do not clear the set of reactions.
     */
    std::string error(error_mode_t mode = LAST_REACTION) const;

    /**
     * Returns a verdict for a set of reactions having occured starting from the previous call. 
     * Clears the set of reactions to be analyzed.
     */
    bool verdict();

    /**
     * Returns a list of reactions.
     * @depricated
     */
    const std::vector<Reaction>& getReactions() const;

    typedef std::list<PointEvent> impl_reactions_t;

protected:

    EventHistory _history;

    impl_reactions_t _implReactions;

////////////////////////////////////////////////////////////////////////////////
// Model Events Management (for Temporal Sequence Recognition)
////////////////////////////////////////////////////////////////////////////////

public:

    void raise(const sequence::ModelEvent &event) { _model_event_manager.add(event); }
    bool check(const sequence::ModelEvent &event) const { return _model_event_manager.contains(event); }

protected:

    sequence::ModelEventManager _model_event_manager;

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

// IMPORTANT: Should be public.
public:
#if CPPTESKHW_CHECK_VERSION(111010)
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::EmptyMessage, CPPTESK_TRIVIAL_ARBITER);
#else
    /// Pseudo interface for reactions that are not assosiated with existing interfaces.
    CPPTESK_DECLARE_INOUT(CPPTESK_DEFAULT_IFACE);
#endif

protected:
    /// Processes a stimulus: serializes a message and applies it to an input interface.
    CPPTESK_DECLARE_PROCESS(onStimulus);
    /// Processes a reaction: detects a reaction, deserializes a message from an output interface and calls a comparator.
    CPPTESK_DECLARE_PROCESS(onReaction);
    /// Processes two messages and registers reactions with errors.
    virtual void reviewMessages(::cpptesk::hw::Process &process,  ::cpptesk::hw::Interface &iface,
                                ::cpptesk::hw::Message &spec_msg, ::cpptesk::hw::Message &impl_msg, time_t &stop_time);

public:
    /// Sets the reaction waiting time.
    void setReactionTimeout(time_t timeout) { _reaction_timeout = timeout; }
    /// Returns the reaction waiting time.
    time_t getReactionTimeout(::cpptesk::hw::Interface &iface) const
        { return iface.isReactionTimeoutValid() ? iface.getReactionTimeout() : _reaction_timeout; }
    /// Sets the implementation reaction waiting time (it is used when the implementation reaction is received earlier than the model reaction).
    void setUnexpectedReactionTimeout(time_t timeout) { _unexpected_reaction_timeout = timeout; }
    /// Returns the implementation reaction waiting time (it is used when the implementation reaction is received earlier than the model reaction).
    time_t getUnexpectedReactionTimeout(::cpptesk::hw::Interface &iface) const
        { return iface.isUnexpectedReactionTimeoutValid() ? iface.getUnexpectedReactionTimeout() : _unexpected_reaction_timeout; }
    /// To maintain compatibility with test systems made earlier.
    time_t getReactionTimeout() const { return _reaction_timeout; }
    time_t getUnexpectedReactionTimeout() const { return _unexpected_reaction_timeout; }

public:

    CPPTESK_DECLARE_PROCESS(listener);
    CPPTESK_DECLARE_PROCESS(singleListener);

////////////////////////////////////////////////////////////////////////////////
// Callback methods (they are not constant to allow users to use side-effect)
////////////////////////////////////////////////////////////////////////////////
protected:
    /// Precondition violation callback.
    virtual void onPreconditionViolation(Process &process);
    /// Postcondition violation callback.
    virtual void onPostconditionViolation(Process &process);
    /// Timeout violation callback.
    virtual void onTimeoutViolation(Process &process);
    /// Normal completion callback.
    virtual void onNormalCompletion(Process &process);
    /// Every cycle callback.
    virtual void onEveryCycle(Process &process);

    /// Every cycle callbacks.
    virtual void onCycleBegin() {}
    virtual void onEveryCycle() {}
    virtual void onCycleEnd()   {}
    /// Invariant checking callback.
    virtual bool invariant()  { return true; }

    /// User condition for forced termination of simulation.
    virtual bool isExitCondition() { return false; }

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods (most of them are virtual, because they can be
// overloaded in a model mediator (adapter); the typical use-case, however, is
// to implement them in a synchronizer class not in a mediator directly)
////////////////////////////////////////////////////////////////////////////////
public:
    /// Initializes simulation using command line parameters.
    void initialize(int argc, char **argv);

    /// Initializes simulation.
    virtual void initialize();
    /// Finalizes simulation.
    virtual void finalize();
    /// Inputs synchronizer: sets input signals of a design.
    virtual void setInputs() { if(_synchronizer) { _synchronizer->setInputs(); } }
    /// Outputs synchronizer: gets output signals of a design.
    virtual void getOutputs() { if(_synchronizer) { _synchronizer->getOutputs(); } }
    /// Thread synchronizer: synchronizes execution threads of a model and simulator.
    virtual void simulate() {  if(_synchronizer) { _synchronizer->simulate(); } }
    /// Registers a synchronizer object.
    void setSynchronizer(Synchronizer *synchronizer) { _synchronizer = synchronizer; }

////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////
private:
    static void launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message);

    struct LaunchParameters
    {
        LaunchParameters(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message):
            model(model), stimulus(stimulus), process(process), iface(iface), message(message) {}

        UserModel *model;
        stimulus_t stimulus;
        Process *process;
        Interface *iface;
        Message *message;
    };

    static void _launch_wrapper(int params_addr_hi, int params_addr_lo);

// IMPORTANT: Should be public (for CYCLE_OF, DELAY_OF, etc.)
public:
    Process* fork(Interface &iface, Message &message, bool clone_message = true);
    void resume(Process *process);
    void suspend(Process *process);

protected:
////////////////////////////////////////////////////////////////////////////////
// Data
////////////////////////////////////////////////////////////////////////////////

    typedef std::vector<Component*> components_list_t;
    components_list_t _components;

    class ProcessWrapper
    {
    public:
        ProcessWrapper(model_t &model, process_t process, const std::string name):
            _model(model), _process(process), _name(name) {}

        /// Start the adapter sequentially
        CPPTESK_DECLARE_NON_VIRTUAL_PROCESS(call)
        {
            // Since it is an internal call, it is transparent to an user
            _model._call = process.call_stack_top(); _model._call.name = _name;
            _model.start(_process, CPPTESK_GET_PARAMETERS());
        }

        /// Start the adapter in a new process
        pid_t call_parallel(Interface &iface, Message &message)
        {
            // Hide the call
            _model._call.reset();
            return _model.start(_process, iface, message, (START_IMMEDIATELY | START_DONT_CLONE_MESSAGE));
        }

    private:
        model_t &_model;
        process_t _process;
        std::string _name;
    };

    typedef std::map<Interface*, ProcessWrapper*> adapters_map_t;
    adapters_map_t _adapters;

    cycle_t _reaction_timeout;
    cycle_t _unexpected_reaction_timeout;

    pid_t _new_pid;

    cycle_t _clock;

    typedef std::map<pid_t, Process*> process_table_t;
    process_table_t _process_table;

    typedef std::vector<Process*> process_list_t;
    process_list_t _active_processes;

    Process *_main_process;

public:

    /// The main loop of a test system.
    Process &main() { return *_main_process; }

public:
    // IMPORTANT: The process stack should be public.
    process_stack_t _process_stack;

    // TODO:
    std::vector<Process::PointDescriptor> getPipelineSlice() const
    {
        std::vector<Process::PointDescriptor> slice(_process_table.size());

        for(process_table_t::const_iterator i = _process_table.begin(); i != _process_table.end(); i++)
            { slice.push_back(i->second->point);  }

        return slice;
    }

////////////////////////////////////////////////////////////////////////////////
// Simulator Synchronizer
////////////////////////////////////////////////////////////////////////////////
private:

    Synchronizer *_synchronizer;

////////////////////////////////////////////////////////////////////////////////
// Command-Line Parameters
////////////////////////////////////////////////////////////////////////////////
private:

    ModelParameters _parameters;
};

template<class UserModel, class UserModelName>
Model<UserModel, UserModelName>::Model():
    CPPTESK_DEFAULT_IFACE("<default>"),
    _reaction_timeout(DEFAULT_REACTION_TIMEOUT),
    _unexpected_reaction_timeout(DEFAULT_UNEXPECTED_REACTION_TIMEOUT),
    _new_pid(0),
    _clock(0),
    _synchronizer(0)
{
    // Register the default interface
    CPPTESK_ADD_IFACE(CPPTESK_DEFAULT_IFACE);

    // The main process of the model
    _main_process = new Process(-1, CPPTESK_DEFAULT_IFACE, CPPTESK_DEFAULT_MESSAGE, time());
    CPPTESK_ASSERTION(_main_process, "No memory to create a process");

    // Pseudo call of the main process.
    _main_process->prepare_call(Process::CallDescriptor("<cpptesk_main_process>"));
    _main_process->start_call();

    // It is a bottom of the process stack
    _process_stack.push_back(_main_process);
}

template<class UserModel, class UserModelName>
Model<UserModel, UserModelName>::~Model()
{
    // Reset the model's pipeline
    reset();

    // Delete the interface adapters
    for(typename adapters_map_t::iterator i = _adapters.begin(); i != _adapters.end(); i++)
        { delete i->second; }
    _adapters.clear();

    // Delete the synchronizer
    if(_synchronizer)
        { delete _synchronizer; }

    // Delete the main process
    if(_main_process)
    {
        // Finish pseudo call of the main process
        _main_process->stop_call();
        // Delete the main process
        delete _main_process;
    }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerComponent(const Component &component)
{
    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Register component " << component.getName());
    _components.push_back(const_cast<Component*>(&component));
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerInterfaceAdapter(Interface &iface, process_t adapter, const std::string &name)
{
    registerInterfaceAdapter(iface, dynamic_cast<model_t&>(*this), adapter, name);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerInterfaceAdapter(Interface &iface, model_t &model, process_t adapter, const std::string &name)
{
    typename adapters_map_t::iterator i = _adapters.find(&iface);

    // Deallocate the previously registered adapter
    if(i != _adapters.end() && !(i->second))
        { delete i->second; }	

    // Register the new one
    _adapters[&iface] = new ProcessWrapper(model, adapter, name);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::launch(UserModel *model, stimulus_t stimulus, Process *process, Interface *iface, Message *message)
{
    CPPTESK_DEBUG_PRINTF_FILE_LINE(DEBUG_TOOL, "Model::launch: model=%p, process=%p, iface=%p, message=%p", model, process, iface, message);

    CPPTESK_ASSERTION(process != 0, "Parameter process is null");
    CPPTESK_ASSERTION(iface   != 0, "Parameter iface is null");
    CPPTESK_ASSERTION(message != 0, "Parameter message is null");

    (model->*stimulus)(*process, *iface, *message);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::_launch_wrapper(int params_addr_hi, int params_addr_lo)
{
    uint64_t addr = ((uint64_t)params_addr_hi << 32) | ((uint64_t)params_addr_lo & 0xffffFFFFull);
    LaunchParameters *params = (LaunchParameters*)addr;

    CPPTESK_ASSERTION(params != 0, "Pointer to parameters is null");

    launch(params->model, params->stimulus, params->process, params->iface, params->message);

    delete params;
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, Message &message, int mode)
{
    Process *process = fork(iface, message, (mode & START_DONT_CLONE_MESSAGE) == 0);
    CPPTESK_ASSERTION(process != NULL, "Process has not forked");

    // Set a call descriptor
    process->prepare_call(get_call()); reset_call();

    priority_t priority = (priority_t)(mode & __CPPTESK_PRIORITY_MASK); 
    process->setPriority(priority == NULL_PRIORITY ? NORMAL_PRIORITY : priority);

    mode &= ~__CPPTESK_PRIORITY_MASK;

    if((mode & START_REQUEST_IFACE) != 0 && iface.isInput())
        { iface.request(*process, *(process->input_message)); }

    if((mode & START_DONT_UPDATE_IFACE_COVERAGE) == 0)
    {
        // Update interface coverage
        coverageTracker << __CPPTESK_IFACE_COVERAGE(&iface);
    }

    Interface *input_iface = process->input_iface;   // The same interface
    Message *input_message = process->input_message; // The message's clone (if START_DONT_CLONE_MESSAGE is not set)

    // If a process finishes unexpectedly, the execution switches to the main loop
    process->ucp->uc_link = _main_process->ucp;

    // The structure is destroyed in _launch_wrapper
    LaunchParameters *params = new LaunchParameters((UserModel*)this, stimulus, process, input_iface, input_message);

    // The makecontext() function supports only integer arguments
    int params_addr_hi = (int)(((uint64_t)params) >> 32);
    int params_addr_lo = (int)(((uint64_t)params) & 0xffffFFFFull);

    makecontext(process->ucp, (void(*)(void))Model<UserModel, UserModelName>::_launch_wrapper, 2, params_addr_hi, params_addr_lo);

    // Execute the first stage of the process within the current cycle if START_IMMEDIATELY is set
    // TODO
    if((mode & START_IMMEDIATELY) != 0)
    {
        Process *this_process = _process_stack.back();

        resume(process);

        // There is no need to check the execution flow consistency, because
        // if a process finishes unexpectedly, the code below is not executed.
        CPPTESK_ASSERTION(_process_stack.back() == this_process, "Process " << process <<
            " has unexpectedly finished (probably, STOP_PROCESS() or STOP_STIMULUS() are missing)");
    }

    if((mode & START_THIS_CYCLE) != 0)
    {
        _active_processes.push_back(process);

        // Sort processes in order of decreasing priorities
        stable_sort(_active_processes.begin(), _active_processes.end(), greaterProcessPriority);
    }

    // Return PID of the newly created process
    return process->pid;
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Interface &iface, int mode)
{
    if((mode & START_DONT_UPDATE_IFACE_COVERAGE) == 0)
    {
        // Update interface coverage
        coverageTracker << __CPPTESK_IFACE_COVERAGE(&iface);
    }

    return start(stimulus, iface, CPPTESK_DEFAULT_MESSAGE, mode);
}

template<class UserModel, class UserModelName>
pid_t Model<UserModel, UserModelName>::start(stimulus_t stimulus, Process &process, Interface &iface, Message &message)
{
    // Set a call descriptor
    process.prepare_call(get_call()); reset_call();

    ((UserModel*)this->*stimulus)(process, iface, message);

    return process.pid;
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::stop(pid_t pid)
{
    process_table_t::iterator i = _process_table.find(pid);

    if(i != _process_table.end())
    {
        delete i->second;
        _process_table.erase(i);
    }
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::isStarted(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isStarted();
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::isWaiting(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isWaiting();
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::isActive(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);
    CPPTESK_ASSERTION(i != _process_table.end(), "Process has not been found");

    return i->second->isActive();
}

template<class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::isFinished(pid_t pid) const
{
    process_table_t::const_iterator i = _process_table.find(pid);

    return i == _process_table.end();
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::cycle()
{
    std::vector<pid_t> to_kill;

    // Forced termination of simulation
    if(_parameters.exit_at_cycle && _clock >= _parameters.exit_at_cycle_value)
        { aspectrace_finishTracing(); ::exit(0); }
    if(_parameters.exit_if_condition && isExitCondition())
        { aspectrace_finishTracing(); ::exit(0); }

    // Callback
    onCycleBegin();
    onEveryCycle();

    // Check a model's invariant
    CPPTESK_ASSERTION(invariant(), "Model invariant violation");

    // Process the model event manager (reset the set of events)
    _model_event_manager.cycle();

    // Process components
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
        { (*i)->cycle(); }

    // Compose the initial set of active processes
    //CPPTESK_ASSERTION(_active_processes.empty(), "Set of active processes is not empty");
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
    {
        Process *process = i->second;

        // If a process has been already executed during the currect cycle (for example, if it
        // was called in onCycleBegin() or onEveryCycle()), then it is not added into the set
        if(process->getTimeMark() == -1u || process->getTimeMark() < time())
            { _active_processes.push_back(process); }
    }

    // Sort processes in order of decreasing priorities
    stable_sort(_active_processes.begin(), _active_processes.end(), greaterProcessPriority);

    while(!_active_processes.empty())
    {
        Process *process = _active_processes.front();

        // The process status after starting is usually EXECUTING, but there are some exceptions
        if(process->status == EXECUTING)
        {
            // Execute a single stage of the process
            resume(process);
            // For dumping the process stack if the assertion below is violated
            __CPPTESK_INSIDE_PROCESS(_process_stack);
            // Checking execution flow consistency
            CPPTESK_ASSERTION(_process_stack.back() == _main_process, "Process " << *process <<
                " has unexpectedly finished (probably, STOP_PROCESS() or STOP_STIMULUS() are missing)");
        }

        if(process->status != EXECUTING)
            { to_kill.push_back(process->pid); }

        // Remove the executed processes from the set of active processes
        //_active_processes.remove(process);
        for(process_list_t::iterator i = _active_processes.begin(); i != _active_processes.end(); i++)
        {
            if(process == *i)
            {
                _active_processes.erase(i);
                break;
            }
        }
    }

    for(std::vector<pid_t>::iterator j = to_kill.begin(); j != to_kill.end(); j++)
        { stop(*j); }

    // Set input signals
    setInputs();

    // Simulate a cycle
    simulate(); onCycleEnd(); _clock++;

    // Get output signals
    getOutputs();
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::reset()
{
    for(process_table_t::iterator i = _process_table.begin(); i != _process_table.end(); i++)
        { delete i->second; }

    _process_table.clear();
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::delay(cycle_t n)
{
    for(cycle_t i = 0; i < n; i++)
        { cycle(); }
}

template<class UserModel, class UserModelName>
const std::vector<Reaction>& Model<UserModel, UserModelName>::getReactions() const
{
    return _history.getReactions();
}

////////////////////////////////////////////////////////////////////////////////
// Working with input and output interfaces
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::onStimulus(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:START]" << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // If a model does not have interface adapters, then do nothing
    if(_adapters.empty())
    {
        // Do nothing
    }
    else
    {
        // Otherwise, try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&CPPTESK_GET_IFACE());

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found for interface " << CPPTESK_GET_IFACE().getName());
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        // Call the serializer having been found
        (i->second->call)(CPPTESK_GET_PARAMETERS());

        // Release the interface if it has not been freed
        if(!CPPTESK_GET_IFACE().isReady())
        {
	    CPPTESK_DEBUG_PRINT(DEBUG_WARN, "Interface " << CPPTESK_GET_IFACE().getName() << " has not been released");
            CPPTESK_RELEASE_IFACE();
        }
    }

    time_t stop_time = time();

    Stimulus stimulus(CPPTESK_GET_IFACE(), start_time, stop_time);
    registerStimulus(stimulus);

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "[STIMULUS:STOP]" << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::onReaction(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    time_t start_time = time();
    time_t stop_time  = time();

    const char *REACTION_START = CPPTESK_GET_PROCESS().listener ? "[LISTENER:START]" : "[REACTION:START]";
    const char *REACTION_STOP  = CPPTESK_GET_PROCESS().listener ? "[LISTENER:STOP]"  : "[REACTION:STOP]";

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, REACTION_START << __CPPTESK_COMMON_START_INFO(start_time) << ": " << CPPTESK_GET_MESSAGE());

    // Emulate the data transmission
    CPPTESK_SET_PRIORITY(CPPTESK_GET_MESSAGE().getPriority());

    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_HOLD)  != 0)
        { CPPTESK_WAIT_PROCESS_FINISHED(CPPTESK_GET_MESSAGE().getPID()); }
    if((CPPTESK_GET_MESSAGE().getInterlock() & LOCK_DELAY) != 0)
        { CPPTESK_DELAY(CPPTESK_GET_MESSAGE().getDelay()); }

    // If a model does not have interface adapters,
    // then register a normal (correct) reaction
    if(_adapters.empty())
    {
        Reaction reaction(NORMAL_REACTION, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), start_time, stop_time, "");
        CPPTESK_DEBUG_PRINT(DEBUG_INFO, reaction);
        registerReaction(reaction);
    }
    else
    {
#ifndef CPPTESK_CLONE_INPUT_MESSAGE
        // The default handler requies clonning of input messages
        CPPTESK_ASSERTION(false, "Default reaction handler does not support non-cloning messages");
#endif
        // Otherwise (there are interface adapters), try to find an appropriate one
        typename adapters_map_t::iterator i = _adapters.find(&CPPTESK_GET_IFACE());

        // Ensure that an interface adapter has been found
        CPPTESK_ASSERTION(i != _adapters.end(), "Adapter has not been found for interface " << CPPTESK_GET_IFACE().getName());
        CPPTESK_ASSERTION(i->second != 0, "Adapter is null");

        Message &spec_msg =  (CPPTESK_GET_MESSAGE());
        Message &impl_msg = *(CPPTESK_GET_MESSAGE().clone());

        // Invalidate the message's fields to be able to check the deserializer
        impl_msg.invalidateFields();
        impl_msg.setUnexpected(false);
        // Call the deserializer having been found
        (i->second->call)(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), impl_msg /* input/output */);

        reviewMessages(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), spec_msg, impl_msg, stop_time);

        // Delete the clone
        delete &impl_msg;
    }

    CPPTESK_DEBUG_PRINT(DEBUG_INFO, REACTION_STOP << __CPPTESK_COMMON_STOP_INFO(start_time, stop_time));

    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::reviewMessages(::cpptesk::hw::Process &process,  ::cpptesk::hw::Interface &iface,
                                                                ::cpptesk::hw::Message &spec_msg, ::cpptesk::hw::Message &impl_msg,
                                                                time_t &stop_time)
{
    std::stringstream buffer;
    bool valid_reaction = true;
    error_type_t error = NORMAL_REACTION;
    time_t reaction_start_time = time();
    stop_time = time();

    if(CPPTESK_GET_PROCESS().listener)
    {
        if(impl_msg.isUnexpected() && !(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & ::cpptesk::hw::UNEXPECTED_REACTION))
        {
            reaction_start_time = time() - 1;
            stop_time = time() - 1;
            error = ::cpptesk::hw::UNEXPECTED_REACTION;
            buffer << "Unexpected reaction: " << impl_msg;
        }
        else
        {
            // Normal reactions and masked reactions from listeners are skipped
            valid_reaction = false;
        }
     }
     else if(CPPTESK_GET_PROCESS().timeout)
     {
         if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & MISSING_REACTION))
         {
             error = MISSING_REACTION;
             buffer << "Missing reaction: " << spec_msg;
         }
         else
         {
            // Masked reactions are skipped
            valid_reaction = false;
         }
     }
     else
     {
         std::string info = spec_msg.compare(impl_msg);

         if(!info.empty())
         {
             if(!(CPPTESK_GET_IFACE().getFailureIgnoreTypes() & INCORRECT_REACTION))
             {
                 error = INCORRECT_REACTION;
                 buffer << "Incorrect reaction: " << info;
             }
             else
             {
                // Masked reactions are skipped
                valid_reaction = false;
             }
         }
     }

     if(error != NORMAL_REACTION)
     {
         std::string explanation = spec_msg.getExplanation();

         if(!explanation.empty())
             { buffer << " " << explanation; }
     }

     if(valid_reaction)
     {
         // Register the reaction
         Reaction reaction(error, CPPTESK_GET_IFACE(), spec_msg, impl_msg, reaction_start_time, stop_time, buffer.str());

         CPPTESK_DEBUG_PRINT((error != NORMAL_REACTION ? DEBUG_ERROR : DEBUG_INFO), reaction);

         if(error != MISSING_REACTION)
             { CPPTESK_DEBUG_PRINT(DEBUG_INFO, impl_msg); }

         registerReaction(reaction);
     }
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::listener(CPPTESK_PROCESS_SIGNATURE())
{
    cycle_t start_time, stop_time;

    CPPTESK_START_PROCESS();

    CPPTESK_GET_PROCESS().listener = true;
    CPPTESK_GET_MESSAGE().setDelay(0);

    while(true)
    {
        start_time = time();

        prepare_call(::cpptesk::hw::Process::CallDescriptor("<cpptesk_unexpected_reaction_listener>"));
        start(&Model<UserModel, UserModelName>::onReaction, CPPTESK_GET_PROCESS(), iface, message);

        stop_time = time();

        if(start_time == stop_time)
            { CPPTESK_CYCLE(); }
    }

    CPPTESK_CYCLE();

    CPPTESK_STOP_PROCESS();
}

template<class UserModel, class UserModelName>
CPPTESK_PROCESS Model<UserModel, UserModelName>::singleListener(CPPTESK_PROCESS_SIGNATURE())
{
    CPPTESK_START_PROCESS();

    // If during a cycle the reaction is not taken, it is supposed to be unexpected
    CPPTESK_CYCLE();

    if(!CPPTESK_GET_IFACE().isReactionTaken())
    {
        time_t stop_time = time();
        PointEvent implReaction = PointEvent(CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), stop_time);
        _implReactions.push_back(implReaction);

        CPPTESK_DELAY(getUnexpectedReactionTimeout(CPPTESK_GET_IFACE()));

        if(getUnexpectedReactionTimeout(CPPTESK_GET_IFACE()) == 0)
        {
            CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Unexpected reaction has been detected at cycle="
                << std::dec << (time() - 1) << std::hex << ": " << CPPTESK_GET_MESSAGE());

            // Increment number of reactions
            CPPTESK_GET_IFACE().incrementEventNumber();
            CPPTESK_GET_MESSAGE().setUnexpected(true);
        }
        else
        {
            for(impl_reactions_t::iterator i = _implReactions.begin(); i != _implReactions.end(); i++)
            {
                if(implReaction.what()->forceCompare(*(i->what())) == Message::COMPARE_OK
                        /*&& time() >= i->when() + getUnexpectedReactionTimeout(iface) - 1*/)
                {
                    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Unexpected reaction has been detected at cycle=" << std::dec << (i->when() - 1) <<
                                                    "; its unexpecting was proven " << getUnexpectedReactionTimeout(CPPTESK_GET_IFACE()) <<
                                                    " cycles later: " << std::hex << ": " << *(i->what()));

                    CPPTESK_GET_PROCESS().listener = true;
                    // Increment number of reactions
                    CPPTESK_GET_IFACE().incrementEventNumber();
                    (*(Message*)(implReaction.what())).setUnexpected(true);

                    reviewMessages(CPPTESK_GET_PROCESS(), CPPTESK_GET_IFACE(), *(Message*)(i->what()), *(Message*)(implReaction.what()), stop_time);

                    break;
                }
            }
        }
    }

    CPPTESK_STOP_PROCESS();
}

////////////////////////////////////////////////////////////////////////////////
// Stimuli & Reactions
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerStimulus(const Stimulus &stimulus)
{
    _history.registerStimulus(stimulus);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerReaction(const Reaction &reaction)
{
    _history.registerReaction(reaction);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerStimulus(const Interface &iface)
{
    Stimulus stimulus(iface, time(), time());
    registerStimulus(stimulus);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::registerReaction(error_type_t error, const Interface &iface, const std::string &info, unsigned type)
{
    Reaction reaction(error, iface, time(), time(), info, type);
    registerReaction(reaction);
}

template <class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::clearReactions()
{
    _history.clearReactions();
}

template <class UserModel, class UserModelName>
std::string Model<UserModel, UserModelName>::error(error_mode_t mode) const
{
    return _history.error(mode);
}

/**
 * Returns a verdict for a set of reactions having been occured starting from the previous call. 
 */
template <class UserModel, class UserModelName>
bool Model<UserModel, UserModelName>::verdict()
{
    return _history.verdict();
}

////////////////////////////////////////////////////////////////////////////////
// Callbacks
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onPreconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.point.file << ":" << process.point.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onPostconditionViolation(Process &process)
{
    std::stringstream buffer;
    buffer << "(" << process.point.file << ":" << process.point.line << ")";
    registerReaction(ASSERTION_VIOLATION, CPPTESK_DEFAULT_IFACE, buffer.str());
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onTimeoutViolation(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onNormalCompletion(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::onEveryCycle(Process &process)
{
    CPPTESK_UNUSED_ARG(process);
}

////////////////////////////////////////////////////////////////////////////////
// Synchronization methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::initialize(int argc, char **argv)
{
    initialize();

    _parameters.init(argc, argv);
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::initialize()
{
    if(_synchronizer) { _synchronizer->initialize(); }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::finalize()
{
    if(_synchronizer) { _synchronizer->finalize(); }

    // Register missing reactions
    for(components_list_t::iterator i = _components.begin(); i != _components.end(); i++)
    {
        Interface *iface = dynamic_cast<Interface*>(*i);

        if(iface != 0 && iface->isOutput())
        {
            Arbiter *arbiter = iface->getArbiter();

            if(arbiter != 0)
            {
                // Get specification reactions for which implementation reactions are expected
                for(requests_t::iterator i = arbiter->begin(); i != arbiter->end(); i++)
                {
                    Reaction reaction(MISSING_REACTION, *iface, *(i->message), time(), time(), "");
                    registerReaction(reaction);
                }
            }
        }
    }

    // Append all reactions to the history
    _history.clearReactions();
    // Print statistics on stimuli and reactions
    _history.traceStatistics();
    // Print diagnosis
    _history.traceDiagnosis();

    // Print simulation time
    CPPTESK_DEBUG_PRINT(DEBUG_INFO, "Simulation time: " << std::dec << time() << " (cycles)");
}

////////////////////////////////////////////////////////////////////////////////
// Auxiliary methods
////////////////////////////////////////////////////////////////////////////////

template<class UserModel, class UserModelName>
Process* Model<UserModel, UserModelName>::fork(Interface &iface, Message &message, bool clone_message)
{
    // IMPORTANT: A newly created process should have a greater PID than the
    // previously launched ones. This guarantees that reset of a control signal
    // (strobe) is done before a new process set the signal.
    Process *process = new Process(_new_pid++, iface, message, time(), clone_message);
    CPPTESK_ASSERTION(process, "No memory");

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::fork: create " << *process);

    _process_table.insert(std::pair<pid_t, Process*>(process->pid, process));

    return process;
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::resume(Process *process)
{
    Process *from;
    Process *jump;

    bool new_process = !process->isStarted();

    CPPTESK_ASSERTION(process, "Parameter process is null");

    // Update timers
    process->cycle(time());

    CPPTESK_ASSERTION(!_process_stack.empty(), "Execution flow inconsistency: empty process stack");

    from = _process_stack.back();
    _process_stack.push_back(process);
    jump = _process_stack.back();

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::resume: switch from " << *from << " to " << *jump);

    swapcontext(from->ucp, jump->ucp);

    if(new_process && process->status != EXECUTING)
    {
        CPPTESK_DEBUG_PRINT(DEBUG_WARN, "Process " << *process
            << " has finished within one cycle (probably, CYCLE() or DELAY() are missing)");
    }

    // Process callbacks
    switch(process->status)
    {
    case UNPREDICT: { onPreconditionViolation  (*process); break; }
    case VIOLATION: { onPostconditionViolation (*process); break; }
    case COMPLETED: { onNormalCompletion       (*process); break; }
    case EXECUTING: { onEveryCycle             (*process); break; }
    default:        { CPPTESK_ASSERTION(false, "Unknown process status"); }
    }
}

template<class UserModel, class UserModelName>
void Model<UserModel, UserModelName>::suspend(Process *process)
{
    Process *from;
    Process *jump;

    CPPTESK_ASSERTION(process != NULL, "Parameter process is null");
    CPPTESK_ASSERTION(_process_stack.size() > 1, "Execution flow inconsistency");

    from = _process_stack.back();
    _process_stack.pop_back();
    jump = _process_stack.back();

    CPPTESK_DEBUG_PRINT_FILE_LINE(DEBUG_TOOL, "Model::suspend: switch from " << *from << " to " << *jump);

    CPPTESK_ASSERTION(from == process, "Execution flow inconsistency");

    swapcontext(from->ucp, jump->ucp);
}

CPPTESK_MODEL(EMPTY_MODEL) {};

inline void set_signal(const std::string &full_name, uint8_t x) { ::vsim_putUInt8Value(full_name.c_str(), x); }
inline void set_value(const std::string &full_name, uint64_t x) { ::vsim_putUInt64Value(full_name.c_str(), x); }
inline uint8_t get_signal(const std::string &full_name) { return ::vsim_getUInt8Value(full_name.c_str()); }
inline uint64_t get_value(const std::string &full_name) { return ::vsim_getUInt64Value(full_name.c_str()); }

}} // namespace cpptesk::hw
