/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Pseudo-random number generator
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>

#include <stdint.h>
#include <iostream>

#define __CPPTESK_GET_RANDOM() \
    ::cpptesk::hw::Random::create()

#define __CPPTESK_ARRAY(type, ...) ((type[]){ __VA_ARGS__ })
#define __CPPTESK_ARRAY_SIZE(type, ...) (sizeof(__CPPTESK_ARRAY(type, __VA_ARGS__))/sizeof(type))

#define CPPTESK_RANDOM(type) \
    ((type)__CPPTESK_GET_RANDOM().rand<type>())

#define CPPTESK_RANDOM_WIDTH(type, width) \
    ((type)__CPPTESK_GET_RANDOM().rand<type>(width))

#define CPPTESK_RANDOM_FIELD(type, lo, hi) \
    ((type)__CPPTESK_GET_RANDOM().rand<type>((unsigned)(lo), (unsigned)(hi)))

#define CPPTESK_RANDOM_BITS(type, lo, hi) \
    CPPTESK_RANDOM_FIELD(type, lo, hi)

#define CPPTESK_RANDOM_RANGE(type, min, max) \
    ((type)__CPPTESK_GET_RANDOM().rand_range<type>((type)(min), (type)(max)))

#define CPPTESK_RANDOM_ENUM(type, min, max) \
    ((type)CPPTESK_RANDOM_RANGE(int, (int)min, (int)max))

#define CPPTESK_RANDOM_CHOICE(type, ...) \
    ((type)__CPPTESK_GET_RANDOM().rand_choice<type>(__CPPTESK_ARRAY(type, __VA_ARGS__), __CPPTESK_ARRAY_SIZE(type, __VA_ARGS__)))

#define CPPTESK_RANDOM_EXCLUDING(type, ...) \
     ((type)__CPPTESK_GET_RANDOM().rand_excluding<type>(__CPPTESK_ARRAY(type, __VA_ARGS__), __CPPTESK_ARRAY_SIZE(type, __VA_ARGS__)))
   
#define CPPTESK_RANDOM_WIDTH_EXCLUDING(type, width, ...) \
     ((type)__CPPTESK_GET_RANDOM().rand_excluding<type>(width, __CPPTESK_ARRAY(type, __VA_ARGS__), __CPPTESK_ARRAY_SIZE(type, __VA_ARGS__)))
   
#define CPPTESK_RANDOM_FIELD_EXCLUDING(type, lo, hi, ...) \
     ((type)__CPPTESK_GET_RANDOM().rand_excluding<type>(lo, hi, __CPPTESK_ARRAY(type, __VA_ARGS__), __CPPTESK_ARRAY_SIZE(type, __VA_ARGS__)))

#define CPPTESK_RANDOM_BITS_EXCLUDING(type, lo, hi, ...) \
    CPPTESK_RANDOM_FIELD_EXCLUDING(type, lo, hi, __VA_ARGS__)

#define CPPTESK_RANDOM_RANGE_EXCLUDING(type, min, max, ...) \
     ((type)__CPPTESK_GET_RANDOM().rand_range_excluding<type>(min, max, __CPPTESK_ARRAY(type, __VA_ARGS__), __CPPTESK_ARRAY_SIZE(type, __VA_ARGS__)))

#define CPPTESK_RANDOM_ENUM_EXCLUDING(type, min, max, ...) \
    ((type)CPPTESK_RANDOM_RANGE_EXCLUDING(int, (int)min, (int)max, __VA_ARGS__))

#define CPPTESK_RANDOM_INDICATOR(probability) \
    (__CPPTESK_GET_RANDOM().rand_indicator(probability))

#define CPPTESK_WITH_PROBABILITY(probability) \
    if(CPPTESK_RANDOM_INDICATOR(probability))

#define CPPTESK_ELSE_WITH_PROBABILITY(probability) \
    else CPPTESK_WITH_PROBABILITY(probability)

#define CPPTESK_ELSE else

#ifdef HW_SHORT_NAMES
    #define RANDOM                 CPPTESK_RANDOM
    #define RANDOM_WIDTH           CPPTESK_RANDOM_WIDTH
    #define RANDOM_FIELD           CPPTESK_RANDOM_FIELD
    #define RANDOM_BITS            CPPTESK_RANDOM_BITS
    #define RANDOM_RANGE           CPPTESK_RANDOM_RANGE
    #define RANDOM_ENUM            CPPTESK_RANDOM_ENUM
    #define RANDOM_CHOICE          CPPTESK_RANDOM_CHOICE
    #define RANDOM_EXCLUDING       CPPTESK_RANDOM_EXCLUDING
    #define RANDOM_WIDTH_EXCLUDING CPPTESK_RANDOM_WIDTH_EXCLUDING
    #define RANDOM_FIELD_EXCLUDING CPPTESK_RANDOM_FIELD_EXCLUDING
    #define RANDOM_BITS_EXCLUDING  CPPTESK_RANDOM_BITS_EXCLUDING
    #define RANDOM_RANGE_EXCLUDING CPPTESK_RANDOM_RANGE_EXCLUDING
    #define RANDOM_ENUM_EXCLUDING  CPPTESK_RANDOM_ENUM_EXCLUDING
    #define RANDOM_INDICATOR       CPPTESK_RANDOM_INDICATOR
    #define WITH_PROBABILITY       CPPTESK_WITH_PROBABILITY
    #define ELSE_WITH_PROBABILITY  CPPTESK_ELSE_WITH_PROBABILITY
    #define ELSE                   CPPTESK_ELSE
#endif // HW_SHORT_NAMES

// Lower bits are more deterministic in linear random generators
#define __CPPTESK_USE_UPPER_BITS(field_width, type_width, value) \
    ((field_width) <= (type_width) - 8 ? (value) >> 8 : (value))

namespace cpptesk {
namespace hw {

class Random
{
public:
    typedef unsigned default_type_t;
    static Random& create();

    template<typename T> static inline T abs(T val)
    {
        return val & ~((T)1 << (8*sizeof(T) - 1));
    }

    template<typename T> static inline T mask(unsigned width)
    {
        return width == 64 ? 0xffffFFFFffffFFFFull : ((T)1 << width) - 1;
    }

    template<typename T> static inline T mask(unsigned lo, unsigned hi)
    {
        return mask<T>((hi - lo) + 1) << lo;
    }

    template<typename T> inline T rand()
    {
        uint64_t res  = __CPPTESK_USE_UPPER_BITS(8*sizeof(T), 8*sizeof(uint32_t), next());

        res &= mask<uint64_t>(8*sizeof(T));

        for(unsigned i = 1; i <= sizeof(T)/4; i++)
            { res <<= 32; res |= next(); }

        return (T)res;
    }

    template<typename T> inline T rand_range(T min, T max)
    {
        T base = abs<T>(rand<T>());

        return min + base % ((max - min) + 1);
    }

    template<typename T> inline T rand_field(unsigned width)
    {
        return __CPPTESK_USE_UPPER_BITS(width, 8*sizeof(T), rand<T>()) & mask<T>(width);
    }

    template<typename T> inline T rand_field(unsigned lo, unsigned hi)
    {
        return __CPPTESK_USE_UPPER_BITS(hi + 1, 8*sizeof(T), rand<T>()) & mask<T>(lo, hi);
    }

    template<typename T> inline T rand_choice(T array[], unsigned size)
    {
        CPPTESK_ASSERTION(size > 0, "Array has zero size");
        return array[rand_range<unsigned>(0u, size - 1)];
    }

    template<typename T> inline T rand_excluding(T array[], unsigned size)
    {
        T result;
	do { result = rand<T>(); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand_range_excluding(T min, T max, T array[], unsigned size)
    {
        T result;
	do { result = rand_range<T>(min, max); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand_field_excluding(unsigned width, T array[], unsigned size)
    {
        T result;
        do { result = rand_field<T>(width); } while(contains<T>(array, size, result));
        return result;
    }

    template<typename T> inline T rand_field_excluding(unsigned lo, unsigned hi, T array[], unsigned size)
    {
        T result;
	do { result = rand_field<T>(lo, hi); } while(contains<T>(array, size, result));
	return result;
    }

    template<typename T> inline T rand(unsigned width)           { return rand_field<T>(width);  }
    template<typename T> inline T rand(unsigned lo, unsigned hi) { return rand_field<T>(lo, hi); }

    template<typename T> inline T rand_excluding(unsigned width, T array[], unsigned size)           { return rand_field_excluding<T>(width, array, size);  }
    template<typename T> inline T rand_excluding(unsigned lo, unsigned hi, T array[], unsigned size) { return rand_field_excluding<T>(lo, hi, array, size); }

    inline bool rand_indicator(float probability)
    {
        CPPTESK_ASSERTION(probability <= 1.0, "Probability is greater than 1");
        return rand_range<unsigned>(0, 99) < (unsigned)(100.0 * probability);
    }

protected:
    Random() {}
    virtual ~Random() {}

    virtual uint32_t next()
    {
        return _value = 1664525 * _value + 1013904223;
    }

    virtual void set_seed(uint32_t seed)
    {
        _value = seed;
    }

    template<typename T> static bool contains(T array[], unsigned size, T item)
    {
        for(unsigned i = 0; i < size; i++)
	{
            if(item == array[i])
                { return true; }
	}

        return false;
    }

private:
    static uint32_t _value;
};

struct Randomizable: public Random
{
    virtual ~Randomizable() {}
    virtual void randomize() = 0;
};

}} // namespace cpptesk::hw

