/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence monitor 
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/sequence/sequence.hpp>
#include <hw/sequence/recognizer.hpp>

#include <hw/unused.hpp>

#include <list>

namespace cpptesk {
namespace hw {
namespace sequence {

/**
 * Basic temporal sequence monitor.
 */
class SequenceMonitor
{
public:
    enum type_t { ONE_TIME, EACH_CYCLE };

    SequenceMonitor(const Sequence &sequence, type_t type = EACH_CYCLE);
    virtual ~SequenceMonitor();

    /// Sequence matching callback.
    virtual void onSuccess(const SequenceRecognizer &recognizer)
        { CPPTESK_UNUSED_ARG(recognizer); }
    /// Sequence mismatching callback.
    virtual void onFailure(const SequenceRecognizer &recognizer)
        { CPPTESK_UNUSED_ARG(recognizer); }

    unsigned getSuccessCount() const { return success_count; }
    unsigned getFailureCount() const { return failure_count; }

    /// Returns number of matches.
    unsigned cycle();

    size_t size() const;

protected:
    typedef std::list<SequenceRecognizer*> recognizers_t;

    /// Sequence being monitored.
    Sequence sequence;
    /// Monitor type.
    type_t type;
    /// Active recognizers of the sequence.
    recognizers_t recognizers;

    unsigned success_count;
    unsigned failure_count;

    /// First cycle flag.
    bool first_cycle;
};

/**
 * Assertion monitor reports an error if the property does not hold.
 */
class AssertionMonitor: public SequenceMonitor
{
public:
    AssertionMonitor(const Sequence &sequence, type_t type = EACH_CYCLE):
        SequenceMonitor(sequence, type) {}

    virtual ~AssertionMonitor() {}

    virtual void onSuccess(const SequenceRecognizer &recognizer)
    {
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_TOOL, "Temporal sequence <" << recognizer.getSequence()
            << "> has been matched at cycle=" << std::dec << recognizer.getTime() << std::hex);
    }

    virtual void onFailure(const SequenceRecognizer &recognizer)
    {
        CPPTESK_ASSERTION(false, "Temporal sequence <" << recognizer.getSequence()
            << "> has been violated");
    }
};

}}} // namespace cpptesk::hw::sequence

