/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Sequence recognizer
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>

#include <hw/sequence/delay.hpp>
#include <hw/sequence/repeat.hpp>
#include <hw/sequence/sequence.hpp>

#include <list>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Sequence Recognizer
////////////////////////////////////////////////////////////////////////////////

class SequenceRecognizer
{
public:
    SequenceRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        sequence(sequence),
        initial_delay(sequence.getDelay()),
        time_constraints(sequence.getTime()),
        i(sequence.begin()),
        parent(parent)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::SequenceRecognizer: " << sequence);
        reset();
    }

    SequenceRecognizer(const Sequence &sequence, const Sequence::const_iterator i, SequenceRecognizer *parent = 0):
        sequence(sequence),
        initial_delay(i->getDelay()),
        time_constraints(i->getTime()),
        i(i),
        parent(parent)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::SequenceRecognizer: " << sequence);
        reset();
    }

    SequenceRecognizer(const Sequence &sequence, const Delay &delay, const Delay &time, SequenceRecognizer *parent = 0):
        sequence(sequence),
        initial_delay(delay),
        time_constraints(time),
        i(sequence.begin()),
        parent(parent)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::SequenceRecognizer: " << sequence);
        reset();
    }

    virtual ~SequenceRecognizer() { reset(); }

    /// The main interface method. Returns a number of new matches for a cycle.
    virtual unsigned cycle();

    bool isWorking()   const { return status == START; }
    bool isFinished()  const { return status == STOP;  }
    bool isSucceeded() const { return status == STOP && count >  0; }
    bool isFailed()    const { return status == STOP && count == 0; }

    /// Returns the sequence being recognized.
    const Sequence& getSequence() const { return sequence; }

    /// Returns the overall number of matches.
    unsigned getMatchCount() const { return count; }

    /// Returns the initial delay range.
    const Delay& getInitialDelay() const { return initial_delay; }

    /// Returns the time constraints.
    const Delay& getTimeConstraints() const { return time_constraints; }

    /// Returns the overall number of cycles having been taken for the sequence acceptance/rejection.
    unsigned getTime() const { return time; }

    /// Checks whether the recognizer is root or not.
    bool isRoot() const { return parent == 0; }
    /// Returns the recognizer level.
    unsigned getLevel() const { return isRoot() ? 0 : parent->getLevel() + 1; }

    /// Returns the number of active child recognizers.
    virtual size_t size() const;

protected:
    /// The main recognition method (should be overloaded in the subclasses).
    /// Returns a number of new matches for a cycle.
    virtual unsigned onCycle();

    /// Starts the recognition process (not used).
    void start() { status = START; }
    /// Stops the recognition process.
    void stop()  { status = STOP; }

    /// Increments the match count.
    unsigned match(unsigned match_count)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::match: " << match_count);

        count += match_count;
        return match_count;
    }

    /// Increments the recognition time.
    void tick()
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SequenceRecognizer::tick");

        time++;
        if(delay.min > 0) { delay.min--; }
        if(delay.max > 0) { delay.max--; }
    }

    /// Resets the recognizer's state (destroys the child recognizers, in particular).
    void reset();

protected:
    /// Data type for recognizer's status.
    enum status_t { STOP, START };

protected:
    /// The recognizer's status (START if it has been started; STOP otherwise).
    status_t status;
    /// The overall number of matches.
    unsigned count;
    /// The sequence being recognized.
    const Sequence &sequence;
    /// Initial delay.
    Delay initial_delay;
    /// Time constraints for recognition.
    Delay time_constraints;
    /// Delay tracker.
    Delay delay;
    /// Recognition time.
    unsigned time;
    /// Delay timeout.
    bool timeout;
    /// Position in a SEQ-sequence.
    Sequence::const_iterator i;
    /// Reference to the parent.
    SequenceRecognizer *parent;

    /// Child recognizers.
    std::list<SequenceRecognizer*> alt;
};

class EventRecognizer: public SequenceRecognizer
{
public:
    EventRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "EventRecognizer::EventRecognizer: " << sequence);
    }

    virtual ~EventRecognizer() {}

    virtual size_t size() const { return 1; }

protected:
    virtual unsigned onCycle();
};

class GuardRecognizer: public SequenceRecognizer
{
public:
    GuardRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent)
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL, "GuardRecognizer::GuardRecognizer: " << sequence);
    }

    virtual ~GuardRecognizer() {}

    virtual size_t size() const { return 1; }

protected:
    virtual unsigned onCycle();
};

class SeqRecognizer: public SequenceRecognizer
{
public:
    SeqRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent),
        main(sequence.isEmpty() ? 0 : new SequenceRecognizer(sequence.getChild(), this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SeqRecognizer::SeqRecognizer: " << sequence << " [" << sequence.getChild() << "]");
    }

    SeqRecognizer(const Sequence &sequence, const Sequence::const_iterator i, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, i, parent),
        main(new SequenceRecognizer(*i, this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "SeqRecognizer::SeqRecognizer: " << sequence << " [" << *i << "]");
    }

    virtual ~SeqRecognizer()
    {
        if(main) { delete main; }

        for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
            { delete *i; }
    }

    virtual size_t size() const
    {
        size_t count = main ? main->size() : 1;

        for(std::list<SequenceRecognizer*>::const_iterator i = alt.begin(); i != alt.end(); i++)
            { count += (*i)->size(); }

        return count;
    }

protected:
    virtual unsigned onCycle();

protected:
    SequenceRecognizer *main;
    std::list<SequenceRecognizer*> alt;
};

class AndRecognizer: public SequenceRecognizer
{
public:
    AndRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent),
        lhs(new SequenceRecognizer(sequence.getLhsChild(), this)),
        rhs(new SequenceRecognizer(sequence.getRhsChild(), this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "AndRecognizer::AndRecognizer: " << sequence);
    }

    virtual ~AndRecognizer()
    {
        if(lhs) { delete lhs; }
        if(rhs) { delete rhs; }
    }

    virtual size_t size() const { return lhs->size() + rhs->size(); }

protected:
    virtual unsigned onCycle();

protected:
    SequenceRecognizer *lhs;
    SequenceRecognizer *rhs;
};

class OrRecognizer: public SequenceRecognizer
{
public:
    OrRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent),
        lhs(new SequenceRecognizer(sequence.getLhsChild(), this)),
        rhs(new SequenceRecognizer(sequence.getRhsChild(), this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "OrRecognizer::OrRecognizer: " << sequence);
    }

    virtual ~OrRecognizer()
    {
        if(lhs) { delete lhs; }
        if(rhs) { delete rhs; }
    }

    virtual size_t size() const { return lhs->size() + rhs->size(); }

protected:
    virtual unsigned onCycle();

protected:
    SequenceRecognizer *lhs;
    SequenceRecognizer *rhs;
};

class RepeatRecognizer: public SequenceRecognizer
{
public:
    RepeatRecognizer(const Sequence &sequence, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent),
        repeat(sequence.getRepeat()),
        main(new SequenceRecognizer(sequence.getChild(), this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "RepeatRecognizer::RepeatRecognizer: " << sequence);
    }

    RepeatRecognizer(const Sequence &sequence, const Repeat &repeat, SequenceRecognizer *parent = 0):
        SequenceRecognizer(sequence, parent),
        repeat(repeat),
        main(new SequenceRecognizer(sequence.getChild(), this))
    {
        CPPTESK_DEBUG_PRINT_FILE_LINE(::cpptesk::hw::DEBUG_TOOL,
            "RepeatRecognizer::RepeatRecognizer: " << sequence);
    }

    virtual ~RepeatRecognizer()
    {
        delete main;

        for(std::list<SequenceRecognizer*>::iterator i = alt.begin(); i != alt.end(); i++)
            { delete *i; }
    }

    virtual size_t size() const
    {
        size_t count = main->size();

        for(std::list<SequenceRecognizer*>::const_iterator i = alt.begin(); i != alt.end(); i++)
            { count += (*i)->size(); }

        return count;
    }

protected:
    virtual unsigned onCycle();

protected:
    Repeat repeat;
    SequenceRecognizer *main;
    std::list<SequenceRecognizer*> alt;
};

}}} // namespace cpptesk::hw::sequence

