/** @file ccp_model.cpp
 *
 * A model of the simple write-invalidate cache coherence protocol (CCP) used to
 * illustrate the C++TESK facilities on cache coherence (CC) verification using
 * the graph exploration (state enumeration) approach. The idea of the example
 * was borrowed from F. Pong and M. Dubois' paper "Verification Techniques for
 * Cache Coherence Protocols" (ACM Computing Surveys (CSUR), Volume 29 Issue 1,
 * March 1997).
 *
 * The CCP model is rather abstract. The goal of the example is to illustrates
 * state abstraction techniques and non-deterministic graph exploration.
 *
 * @author Alexander Kamkin (kamkin@ispras.ru)
 */

#include <ccp_model.h>

#include <algorithm>

namespace cpptesk {
namespace examples {
namespace ccp {

////////////////////////////////////////////////////////////////////////////////
// Cache
////////////////////////////////////////////////////////////////////////////////

#define UNEXPECTED() CPPTESK_ASSERTION(false, "Unexpected stimulus")

// Definition of the cache state machine.
void Cache::process(requester_t r, operation_t o)
{
    switch(_state)
    {
    case INVALID:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { UNEXPECTED();     break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = INVALID; break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            break;
        }
    case SHARED:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { _state = SHARED;  break; }
                }
            }
            break;
        }
    case DIRTY:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = DIRTY;   break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { UNEXPECTED();     break; }
                }
            }
            break;
        }
    default: { UNEXPECTED(); }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Cache Coherence Model
////////////////////////////////////////////////////////////////////////////////

/// Tracks the cache coherence protocol's test coverage.
void CacheCoherenceModel::trackCoverage(unsigned id, Cache::operation_t o)
{
    const std::vector<Processor> &cpu = processors();

    for(std::vector<Processor>::const_iterator i = cpu.begin(); i != cpu.end(); i++)
    {
        coverageTracker << mainCoverage(i->state(), o, (id == i->id() ? Cache::THIS : Cache::OTHER));
    }
}

////////////////////////////////////////////////////////////////////////////////
// State Abstraction
////////////////////////////////////////////////////////////////////////////////

/// Returns the state depending on the abstraction type.
std::string CacheCoherenceModel::getState() const
{
    switch(_state_abstraction)
    {
    case VECTOR:   { return getVectorState(); }
    case MULTISET: { return getMultiSetState(); }
    case SET:      { return getSetState(); }
    }

    return "";
}

/// Returns the state (vector of individual cache states).
std::string CacheCoherenceModel::getVectorState() const
{
    return toString(getStateVector());
}

/// Exploits the vector-to-multiset abstraction.
std::string CacheCoherenceModel::getMultiSetState() const
{
    std::vector<std::string> multiset = getStateVector();
    sort(multiset.begin(), multiset.end());
    return toString(multiset);
}

/// Exploits the vector-to-set abstraction.
std::string CacheCoherenceModel::getSetState() const
{
    std::vector<std::string> set, seq = getStateVector();

    sort(seq.begin(), seq.end());

    for(std::vector<std::string>::const_iterator i = seq.begin(); i != seq.end();)
    {
        std::string item = *i;
        set.push_back(item);
        for(++i; i != seq.end() && (*i) == item; i++);
    }

    return toString(set);
}

std::vector<std::string> CacheCoherenceModel::getStateVector() const
{
    std::vector<std::string> vector;

    for(std::vector<Processor>::const_iterator i = _processors.begin(); i != _processors.end(); i++)
        { vector.push_back(i->getState()); }

    return vector;
}

std::string CacheCoherenceModel::toString(const std::vector<std::string> &state) const
{
    std::stringstream buffer;

    for(std::vector<std::string>::const_iterator i = state.begin(); i != state.end(); i++)
        { buffer << *i; }

    return buffer.str();
}

////////////////////////////////////////////////////////////////////////////////
// Transition Abstraction
////////////////////////////////////////////////////////////////////////////////

/// Returns the number of transitions. 
unsigned CacheCoherenceModel::getTransitionNumber() const
{
    switch(_trans_abstraction)
    {
    case INDEX: { return getIndexNumber(); }
    case KIND:  { return getKindNumber(); }
    }

    return 0;
}

/// Returns the processor identifier or -1.
unsigned CacheCoherenceModel::getProcessorId(unsigned transitionId) const
{
    switch(_trans_abstraction)
    {
    case INDEX: { return getIndexId(transitionId); }
    case KIND:  { return getKindId(transitionId); }
    }

    return -1;
}

/// Returns the number of processors.
unsigned CacheCoherenceModel::getIndexNumber() const
{
    return size();
}

/// Returns the processor identifier.
unsigned CacheCoherenceModel::getIndexId(unsigned transitionId) const
{
    return transitionId;
}

/// Returns the number of transition kinds (cache states).
unsigned CacheCoherenceModel::getKindNumber() const
{
    return 3;
}

/// Returns the processor identifier.
unsigned CacheCoherenceModel::getKindId(unsigned transitionId) const
{
    std::vector<unsigned> transitions;

    for(std::vector<Processor>::const_iterator i = _processors.begin(); i != _processors.end(); i++)
    {
        if((unsigned)i->state() == transitionId)
            { transitions.push_back(i->id()); }
    }

    return transitions.empty() ? -1u : transitions[rand() % transitions.size()];
}



}}} // namespace cpptesk::examples::ccp
