/** @file generator.h
 *
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __GENERATOR_H_INCLUDED__
#define __GENERATOR_H_INCLUDED__

#include <atl/list.h>
#include <atl/map.h>
#include <atl/string.h>
#include <ts/generated.h>
#include <ts/scenario.h>

#ifdef __cplusplus
extern "C" {
#endif

#define NETFSM_ITERATION_CACHE 

#ifdef NETFSM_ITERATION_CACHE_DISABLE
    #undef NETFSM_ITERATION_CACHE
#endif // NETFSM_ITERATION_CACHE_DISABLE

/******************************************************************************/
/**                                Test Engine                               **/
/******************************************************************************/

/******************************************************************************/
/**                       Test Engine / Basic Data Types                     **/
/******************************************************************************/

/** @def TE_UNKNOWN_DEGREE
 * Unknown out-degree of a node.
 */
#define TE_UNKNOWN_DEGREE (-1)

/** @typedef TE_ArcSymbolPair
 * Iteration cache entry.
 */
typedef struct TE_Iteration
{
    /// Iteration order number.
    int index;
    /// Scenario function index.
    int kind;
    /// Iteration context.
    Object *value;
} TE_Iteration;

/** @typedef TE_NodeId
 * A unique node identifier.
 */
typedef struct TE_NodeId
{
    /// Identifier of a node.
    String *id;
    /// Out-degree of a node (unknown degree = -1).
    int deg;
#ifdef NETFSM_ITERATION_CACHE
    /// Internal data of a node.
    struct
    {
        /// Cached iteration.
        TE_Iteration cache;
    } sysdata;
#endif // NETFSM_ITERATION_CACHE
} TE_NodeId;

/** @var type_TE_NodeId
 * A type descriptor for an node identifier.
 */
extern const Type type_TE_NodeId;

/** @fn TE_NodeId* create_NodeId(String *id)
 * Creates a node identifier.
 * @param id the string representing the node identifier.
 */
TE_NodeId* create_NodeId(String *id);

/** @fn TE_NodeId* create_NodeIdWithDegree(String *id)
 * Creates a node identifier with a degree specified.
 * @param id the string representing the node identifier.
 * @param degree the degree of the node.
 */
TE_NodeId* create_NodeIdWithDegree(String *id, int degree);

/** @fn int hash_String(String *str)
 * Calculates a string's hash code.
 * @param str the string (should be r()'ed).
 */
int hash_String(String *str);

/** @fn int hash_NodeId(TE_NodeId *node_id)
 * Calculates a node's hash code.
 * @param node_id the string representing the node identifier (should be r()'ed).
 */
int hash_NodeId(TE_NodeId *node_id);

/** @def NODE_ID_STRING
 * Converts a node identifier into a char*-string.
 */
#define NODE_ID_STRING(node_id) \
    ((node_id) != NULL ? ((node_id->id) != NULL ? toCharArray_String(r(node_id->id)) : "unknown") : "UNKNOWN")

/** @typedef TE_StimulusId
 * A unique stimulus identifier.
 */
typedef struct TE_StimulusId
{
    /// An iteration number.
    int index;
} TE_StimulusId;

/** @fn TE_StimulusId create_StimulusId(int index)
 * Creates a stimulus identifier.
 * @param index the iteration within the given scenario function.
 */
TE_StimulusId create_StimulusId(int index);

/** @fn int compare_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs)
 * Compares two stimulus identifiers. Returns zero iff two identifiers are equal.
 */
int compare_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs);

/** @fn bool equal_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs)
 * Compares two stimulus identifiers. Returns true iff two identifier are equal.
 */
bool equal_StimulusId(TE_StimulusId lhs, TE_StimulusId rhs);

/** @typedef TE_ArcId
 * A unique arc identifier.
 */
typedef struct TE_ArcId
{
    /// A start node identifier.
    TE_NodeId *start_node_id;
    /// A stimulus identifier.
    TE_StimulusId stimulus_id;
    /// An end node identifier.
    TE_NodeId *end_node_id;
} TE_ArcId;

/** @var type_TE_ArcId
 * A type descriptor for an arc identifier.
 */
extern const Type type_TE_ArcId;

/** @fn TE_ArcId* create_ArcId(TE_NodeId *start_node_id, TE_StimulusId stimulus, TE_NodeId *end_node_id)
 * Creates an arc identifier.
 * @param start_node_id the start node identifier.
 * @param stimulus_id the stimulus identifier.
 * @param end_node_id the end node identifier.
 */
TE_ArcId* create_ArcId(TE_NodeId *start_node_id, TE_StimulusId stimulus_id, TE_NodeId *end_node_id);

/** @typedef TE_ArcList
 * A list of arc identifiers.
 */
typedef List TE_ArcList;

/******************************************************************************/
/**                       Test Engine / Graph Data Base                      **/
/******************************************************************************/

/** @typedef TE_GraphStatus
 * A status of a graph data base.
 */
typedef enum TE_GraphStatus
{
    /// A graph data base has not been initialized.
    TE_DISABLED  = -1,
    /// A graph data base has been initialized.
    TE_ENABLED   =  0,
    /// A graph data base has been initialized, and traversal has been started.
    TE_STARTED   =  1,
    /// Traversal has been finished, and a graph data base has been destroyed.
    TE_DESTROYED = -2,
} TE_GraphStatus;

struct TE_GraphMutex;

/** @typedef TE_MutexFunction
 *  A mutex function.
 */
typedef void (*TE_MutexFunction)(struct TE_GraphMutex*);

/** @typedef TE_GraphMutex
 * An interface for a mutex object.
 */
typedef struct TE_GraphMutex
{
    /// A function that initializes a mutex.
    TE_MutexFunction init;
    /// A function that locks a graph data base.
    TE_MutexFunction lock;
    /// A function that unlocks a graph data base.
    TE_MutexFunction unlock;
    /// A function that release a mutex.
    TE_MutexFunction finish;
    /// A mutex handle.
    void *handle;
} TE_GraphMutex;

/** @var TE_NO_SYNC_MUTEX
 * A predefined mutex that does not perform any synchronization.
 */
extern TE_GraphMutex TE_NO_SYNC_MUTEX;

/** @var TE_DEFAULT_MUTEX
 * A default mutex.
 */
extern TE_GraphMutex TE_DEFAULT_MUTEX;

/** @typedef TE_GraphInfo
 * Information on a graph.
 */
typedef struct TE_GraphInfo
{
    /// Traverser identifier.
    int cpu_id;
    /// Number of traversed nodes.
    int nodes_traversed;
    /// Number of nodes traversed locally.
    int nodes_traversed_locally;
    /// Number of finished nodes.
    int nodes_finished;
    /// Number of all known arcs.
    int arcs_known;
    /// Number of nondeterministic arcs.
    int arcs_nondeterministic;
    /// Number of traversed arcs.
    int arcs_traversed;
    /// Number of traversed arcs.
    int arcs_traversed_locally;
    /// Number of traversed loops.
    int loops_traversed;
} TE_GraphInfo;

/** @fn void init_GraphInfo(TE_GraphInfo *graph_info)
 * Initializes graph information.
 */
void init_GraphInfo(TE_GraphInfo *graph_info);

/** @typedef TE_GraphParameters
 * Traversal parameters.
 */
typedef struct TE_GraphParameters
{
    bool print_progress;
} TE_GraphParameters;

/** @fn void init_GraphParameters(TE_GraphParameters *graph_params)
 * Initializes graph parameters.
 */
void init_GraphParameters(TE_GraphParameters *graph_params);

/** @typedef TE_Graph
 * A graph data base.
 */
typedef struct TE_Graph
{
    /// A traversal status.
    TE_GraphStatus status;
    /// Information on a graph.
    TE_GraphInfo info;
    /// Traversal parameters.
    TE_GraphParameters params;
    /// A flag representing a request to pause a test engine.
    bool pause_req;
    /// A flag representing an acknowledgement for the pause request.
    bool pause_ack;
    /// An implementation-specific mutual exclusion object.
    TE_GraphMutex *mutex;
    /// A traversal length.
    int length;
    /// An implementation-specific fatal error code.
    int error;
    /// Number of non-critical failures (e.g., postcondition violations).
    int fails;
    /// A list of arcs having been localy updated (to be given to the synchronizer).
    TE_ArcList *updates;
    /// An implementation-specific test engine state.
    void *data;
} TE_Graph;

/** @typedef TE_UpdateMode
 * An update mode.
 */
typedef enum TE_UpdateMode
{
    /// A graph data base accumulates arcs to be given to the synchronizer.
    TE_UPDATE_ON  = 1,
    /// A graph data base does not accumulate arcs.
    TE_UPDATE_OFF = 0
} TE_UpdateMode;

/** @fn void init_Graph(TE_Graph *graph, TE_GraphMutex *mutex)
 * Initializes a graph data base (does not initialize an implementation-specific part).
 * @param graph the graph data base.
 * @param mode the update mode.
 * @param mutex the mutex object.
 */
void init_Graph(TE_Graph *graph, TE_UpdateMode mode, TE_GraphMutex *mutex);

/** @fn void destroy_Graph(TE_Graph *graph)
 * Destroys a graph data base (does not destroy an implementation-specific part).
 * @param graph the graph data base.
 */
void destroy_Graph(TE_Graph *graph);

/** @fn void lock_Graph(TE_Graph *graph)
 * Locks a graph data base.
 */
void lock_Graph(TE_Graph *graph);

/** @fn void unlock_Graph(TE_Graph *graph)
 * Unlocks a graph data base.
 */
void unlock_Graph(TE_Graph *graph);

/** @fn TE_GraphStatus getStatus_Graph(TE_Graph *graph)
 * Returns a graph data base status.
 * @param graph the graph data base. 
 */
TE_GraphStatus getStatus_Graph(TE_Graph *graph);

/** @fn int getLength_Graph(TE_Graph *graph)
 * Returns a traversal length.
 * @param graph the graph data base. 
 */
int getLength_Graph(TE_Graph *graph);

/** @fn int incLength_Graph(TE_Graph *graph, int path_length)
 * Increments a traversal length.
 * @param graph the graph data base. 
 */
int incLength_Graph(TE_Graph *graph, int path_length);

/** @fn int getFailures_Graph(TE_Graph *graph)
 * Returns a number of non-critical failures.
 * @param graph the graph data base. 
 */
int getFailures_Graph(TE_Graph *graph);

/** @fn int incFailures_Graph(TE_Graph *graph)
 * Increments a number of failures and returns an updated number.
 * @param graph the graph data base. 
 */
int incFailures_Graph(TE_Graph *graph);

/** @fn int getError_Graph(TE_Graph *graph)
 * Returns a fatal error code.
 * @param graph the graph data base. 
 */
int getError_Graph(TE_Graph *graph);

/** @fn void sleep_Graph(TE_Graph *graph)
 * Suspends traversal if a pause request has been set.
 * @param graph the graph data base.
 */
void sleep_Graph(TE_Graph *graph);

/** @fn void printProgress_Graph(TE_Graph *graph)
 * Prints traversal progress.
 * @param graph the graph data base.
 */
void printProgress_Graph(TE_Graph *graph);

/******************************************************************************/
/**                       Test Engine / Main Interface                       **/
/******************************************************************************/

/** @typedef TE_UpdateType
 * An update type.
 */
typedef enum TE_UpdateType
{
    /// Update being done by a synchronizer.
    TE_REMOTE = 0,
    /// Update being done by a traverser.
    TE_LOCAL  = 1
} TE_UpdateType;

/** @typedef TE_UpdateStatus
 * An update status (bitwise mask).
 */
typedef enum TE_UpdateStatus
{
    /// Nothing new has been added.
    TE_OLD_ALL  = 0x0,
    /// A new node has been added.
    TE_NEW_NODE = 0x1,
    /// A new arc has been added.
    TE_NEW_ARC  = 0x2
} TE_UpdateStatus;

/** @typedef TE_Init
 * A test engine initialization function.
 */
typedef void (*TE_Init)(TE_Graph*, int, char**);

/** @typedef TE_Config
 * A test engine configuration function.
 */
typedef void (*TE_Config)(TE_Graph*, const char*, const char*);

/** @typedef TE_Start
 * A test engine function that sets a start node.
 */
typedef TE_UpdateStatus (*TE_Start)(TE_Graph*, TE_NodeId*);

/** @typedef TE_Update
 * A test engine function that updates a graph data base.
 */
typedef TE_UpdateStatus (*TE_Update)(TE_Graph*, TE_UpdateType, TE_ArcId*, TE_NodeId*);

/** @typedef TE_GetPath
 * A test engine function that returns a path to a nonfinished node.
 */
typedef TE_ArcList* (*TE_GetPath)(TE_Graph*, TE_NodeId*);

/** @typedef TE_ReachNode
 * A test engine function that returns a path to a given node from the initial one.
 */
typedef TE_ArcList* (*TE_ReachNode)(TE_Graph*, TE_NodeId*, TE_ArcId*);

/** @typedef TE_GetError
 * A test engine function that returns a string describing an error.
 */
typedef String* (*TE_GetError)(TE_Graph*);

/** @typedef TE_Finish
 * A test engine finalization function.
 */
typedef void (*TE_Finish)(TE_Graph*);

/** @typedef TE_ToFile
 * A test engine function that dumps a checkpoint into a file.
 */
typedef bool (*TE_ToFile)(TE_Graph*, const char*);

/** @typedef TE_FromFile
 * A test engine function that constructs a graph data base from a checkpoint.
 */
typedef bool (*TE_FromFile)(TE_Graph*, const char*);

struct ArcSymbolExecutor;

/** @typedef TestEngine
 * A type describing a test engine.
 */
typedef struct TestEngine
{
    /// A test engine name.
    const char*       name;
    /// An initialization function.
    TE_Init           init;
    /// A configuration function.
    TE_Config         config;
    /// A start function.
    TE_Start          start;
    /// A function that updates a graph data base.
    TE_Update         update;
    /// A function that returns a path to a nonfinished node.
    TE_GetPath        getPath;
    /// A function that returns a path to a given node from the initial one.
    TE_ReachNode      reachNode;
    /// A function that returns an error description.
    TE_GetError       getError;
    /// A finalization function.
    TE_Finish         finish;
    /// A function that dumps a checkpoint into a file.
    TE_ToFile         toFile;
    /// A function that reads a graph data base from a file.
    TE_FromFile       fromFile;
    /// Auxiliary components.
    struct
    {
        /// An arc executor used to enumerate arcs and execute stimuli.
        struct ArcSymbolExecutor *executor;
    } sysdata;
} TestEngine;

/******************************************************************************/
/**                          Test Engine / Wrappers                          **/
/******************************************************************************/

/** @fn const char* getName_TestEngine(TestEngine *engine)
 * Returns a test engine name.
 * @param engine the test engine.
 */
const char* getName_TestEngine(TestEngine *engine);

/** @fn void init_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test, int argc, char **argv)
 * Initializes a graph data base.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param test the test scenario.
 * @param argc the number of command-line parameters.
 * @param argv the command-line parameters.
 */
void init_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test, int argc, char **argv);

/** @fn void start_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
 * Starts traversal (sets a start node for traversal and expands it if the node is new).
 * @param engine the test engine.
 * @param graph the initialized graph data base.
 * @param node_id the start node identifier.
 */
void start_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id);

/** @fn TE_UpdateStatus updateLocal_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
 * Updates a graph data base and returns an update status. New nodes are expanded.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param arc_id the identifier of the arc to be updated.
 * @param new_end_node_id the identifier of the end node (if it is NULL, it is not taken into account).
 */
TE_UpdateStatus updateLocal_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id, TE_NodeId *new_end_node_id);

/** @fn TE_UpdateStatus updateRemote_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id)
 * Updates a graph data base and returens an update status. New nodes are not expanded.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param arc_id the identifier of the arc to be updated.
 */
TE_UpdateStatus updateRemote_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id);

/** @fn void updateRemoteAll_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcList *arc_list)
 * Updates a graph data base with a list of arcs. New nodes are not expanded.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param arc_list the list of arc identifiers.
 */
void updateRemoteAll_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcList *arc_list);

/** @fn TE_ArcList* getUpdates_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Returns a list of localy updated arcs.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
TE_ArcList* getUpdates_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn TE_ArcList* getPath_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
 * Returns a path to a nonfinished node.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param node_id the current node identifier.
 */
TE_ArcList* getPath_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id);

/** @fn TE_ArcList* reachNode_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
 * Returns a path to a given node from the initial one.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param node_id the target node identifier.
 * @param arc_id the arc identifier (it might be NULL).
 */
TE_ArcList* reachNode_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id);

/** @fn TE_ArcList* getPath_TestEngine(TestEngine *engine, TE_Graph *graph, TE_NodeId *node_id)
 * Returns a description of an error.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
String* getError_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn void getInfo_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Returns information on a graph.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
void getInfo_TestEngine(TestEngine *engine, TE_Graph *graph, TE_GraphInfo *info);

/** @fn void finish_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Finishes traversal and destroys a graph data base.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
void finish_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn bool toFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *filename)
 * Dumps a checkpoint into a file.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param filename the file name.
 */
bool toFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *filename);

/** @fn bool fromFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *filename)
 * Reads a checkpoint from a file.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param filename the file name.
 */
bool fromFile_TestEngine(TestEngine *engine, TE_Graph *graph, const char *filename);

/** @fn bool fromFile_Default(TestEngine *engine, TE_Graph *graph, const char *filename)
 * A default function for reading from a file.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param filename the file name.
 */
bool fromFile_Default(TestEngine *engine, TE_Graph *graph, const char *filename);

/** @fn void suspend_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Suspends test engine execution.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
void suspend_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn void resume_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Resumes test engine execution.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
void resume_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn bool isSuspended_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Checks if the test engine has been suspended or not.
 * @param engine the test engine.
 * @param graph the graph data base.
 */
bool isSuspended_TestEngine(TestEngine *engine, TE_Graph *graph);

/******************************************************************************/
/**                        Test Engine / Attributes                          **/
/******************************************************************************/

/** @def TE_ATTR_CPU_ID
 * An attribute corresponding to a CPU identifier (GET/SET).
 * Values: strings representing non-negative decimal numbers.
 */
#define TE_ATTR_CPU_ID "cpu"

/** @def TE_ATTR_TRAVERSED_NODES
 * An attribute corresponding to a number of traversed nodes (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_TRAVERSED_NODES "nodes_traversed"

/** @def TE_ATTR_LOCALLY_TRAVERSED_NODES
 * An attribute corresponding to a number of locally traversed nodes (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_LOCALLY_TRAVERSED_NODES "nodes_traversed_locally"

/** @def TE_ATTR_FINISHED_NODES
 * An attribute corresponding to a number of finished nodes (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_FINISHED_NODES "nodes_finished"

/** @def TE_ATTR_KNOWN_ARCS
 * An attribute corresponding to a number of all known arcs (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_KNOWN_ARCS "arcs_known"

/** @def TE_ATTR_NONDETERM_ARCS
 * An attribute corresponding to a number of nondeterministic arcs (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_NONDETERM_ARCS "arcs_nondeterministic"

/** @def TE_ATTR_TRAVERSED_ARCS
 * An attribute corresponding to a number of traversed arcs (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_TRAVERSED_ARCS "arcs_traversed"

/** @def TE_ATTR_LOCALLY_TRAVERSED_ARCS
 * An attribute corresponding to a number of locally traversed arcs (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_LOCALLY_TRAVERSED_ARCS "arcs_traversed_locally"

/** @def TE_ATTR_TRAVERSED_LOOPS
 * An attribute corresponding to a number of traversed loops (GET).
 * Values: strings representing non-negative decimal numbers or -1 (unknown).
 */
#define TE_ATTR_TRAVERSED_LOOPS "loops_traversed"

/** @def TE_ATTR_STATUS
 * An attribute corresponding to a traversal status (GET).
 * Values:
 *  TE_ATTR_STATUS_DISABLED - a test engine has not been initialized;
 *  TE_ATTR_STATUS_ENABLED  - a test engine has been initialized (traversal is in progress);
 *  TE_ATTR_STATUS_FINISHED - a test has been successfully finished;
 *  TE_ATTR_STATUS_FAILED   - a test has been finished due to errors;
 *  TE_ATTR_STATUS_ABORTED  - a test has been deliberately aborted by a user or a remote component.
 */
#define TE_ATTR_STATUS "status"

#define TE_ATTR_STATUS_DISABLED "disabled"
#define TE_ATTR_STATUS_ENABLED  "enabled"
#define TE_ATTR_STATUS_STARTED  "started"
#define TE_ATTR_STATUS_FINISHED "finished"
#define TE_ATTR_STATUS_FAILED   "failed"
#define TE_ATTR_STATUS_ABORTED  "aborted"

/** @def TE_ATTR_FAILURES
 * An attribute corresponding to a number of non-critical failures (GET).
 * Values: strings representing non-negative decimal numbers.
 */
#define TE_ATTR_FAILURES "failures"

/** @def TE_ATTR_ERROR
 * An attribute corresponding to a description of a fatal error (GET).
 * Values: textual strings.
 */
#define TE_ATTR_ERROR "error"

/** @fn String* getAttrValue_TestEngine(TestEngine *engine, Graph *graph, const char *attr_name)
 * A function implementing an attribute-name-to-attribute-value interface with a coordinator.
 * If an attribute is not supported, the function returns NULL.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param attr_name the attribute name.
 */
String* getAttrValue_TestEngine(TestEngine *engine, TE_Graph *graph, const char *attr_name);

/** @fn Map* getAllAttrValues_TestEngine(TestEngine *engine, Graph *graph)
 * Returns an attribute-name-to-attribute-value map.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param attr_name the attribute name.
 */
Map* getAllAttrValues_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn void setAttrValue_TestEngine(String *attr_name, String *attr_value)
 * Sets value of an attribute.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param attr_name the attribute name.
 * @param attr_value the attribute value.
 */
void setAttrValue_TestEngine(TestEngine *engine, TE_Graph *graph, String *attr_name, String *attr_value);

/** @fn void setAllAttrValues_TestEngine(Map *all_attrs)
 * Sets values for several attributes.
 * @param engine the test engine.
 * @param graph the graph data base.
 * @param all_attrs the name-to-value map.
 */
void setAllAttrValues_TestEngine(TestEngine *engine, TE_Graph *graph, Map *all_attrs);

/******************************************************************************/
/**                       Test Engine / Main Function                        **/
/******************************************************************************/

/** @fn void main_TestEngine(TestEngine *engine, TestScenario *test, int argc, char **argv);
 * A function implementing an active part of traversal.
 * @param engine the test engine (should be initialized).
 * @param graph the graph data base (should be initialized).
 * @param test the test scenario.
 * @param argc the number of command-line parameters.
 * @param argv the command-line parameters.
 */
bool main_TestEngine(TestEngine *engine, TE_Graph *graph, TestScenario *test, int argc, char **argv);

/******************************************************************************/
/**                       Test Engine / User Functions                       **/
/******************************************************************************/

/** @fn void set_TestEngine(TestEngine *engine, TE_Graph *graph)
 * Sets the current test engine and graph data base.
 */
void set_TestEngine(TestEngine *engine, TE_Graph *graph);

/** @fn bool start_TestScenario(int argc, char **argv, TestScenario *test)
 * Starts a test scenario.
 */
bool start_TestScenario(int argc, char **argv, TestScenario *test);

/** @fn int localmain(TestEngine *engine, TestScenario *test, int argc, char **argv)
 * Starts a test on a local machine.
 */
int localmain(TestEngine *engine, TestScenario *test, int argc, char **argv);

#ifdef __cplusplus
} // extern "C" 

#include <ts/scenario.hpp>

namespace cpptesk { 
namespace ts {

int localmain(TestEngine *engine, CppTestScenario *test, int argc, char **argv);

}}

#endif // __cplusplus

#endif

