/*Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Fuzzy Associative Elimination 
//           (Rule # 9 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[y1:x2]+[x1:y2] => [y1:y2]+[x1:x2]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/rule.hpp>
#include <hw/diagnostics/rule/fuzzy_associative.hpp>
#include <hw/diagnostics/transform_unit.hpp>
#include <hw/diagnostics/reaction_set.hpp>

#include <stdio.h>
#include <hw/message.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet FuzzyAssociativeEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_1, ::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_2)    // Collapse the reactions
    {
        ReactionSet _transform_out;
        
        const ::cpptesk::hw::Message *_message1Spec, *_message1Impl;
        const ::cpptesk::hw::Message *_message2Impl, *_message2Spec;
        
        // Message from reaction 1
        _message1Spec = _reactionUnit_1._reaction.getSpecMessage();
        _message1Impl = _reactionUnit_1._reaction.getImplMessage();
        // Message from reaction 2
        _message2Spec = _reactionUnit_2._reaction.getSpecMessage();
        _message2Impl = _reactionUnit_2._reaction.getImplMessage();
        
        // Get iface
        const ::cpptesk::hw::Interface *iface1 = _reactionUnit_1._reaction.where();
        const ::cpptesk::hw::Interface *iface2 = _reactionUnit_2._reaction.where();
        
        // Get "start" and "end" time of reactions
        cycle_t start_time   =  _reactionUnit_1._reaction.whenStarted();
        cycle_t start_time_2 =  _reactionUnit_2._reaction.whenStarted();
        cycle_t end_time     =  _reactionUnit_1._reaction.whenStopped();
        cycle_t end_time_2   =  _reactionUnit_2._reaction.whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        ::cpptesk::hw::Reaction reactionIncorrect1 = Reaction(INCORRECT_REACTION, *iface1, *_message1Spec, *_message2Impl, start_time, end_time, "Incorrect reaction");
        ::cpptesk::hw::Reaction reactionIncorrect2 = Reaction(INCORRECT_REACTION, *iface2, *_message2Spec, *_message1Impl, start_time, end_time, "Incorrect reaction");
        
        _transform_out.addReaction(reactionIncorrect1);
        _transform_out.addReaction(reactionIncorrect2);
        
        // Set flag
        _reactionUnit_1.useInCollapse();
        _reactionUnit_2.useInCollapse();
       
        return _transform_out;
    }

    int FuzzyAssociativeEliminationRule::match(::cpptesk::hw::Reaction _reaction_1, ::cpptesk::hw::Reaction _reaction_2)
    // Compare the reactions
    {
        int _match = FALSE_MATCH;
        int _match11 = FALSE_MATCH, _match22 = FALSE_MATCH, _match12 = FALSE_MATCH, _match21 = FALSE_MATCH;

        if((_reaction_1.getError() == INCORRECT_REACTION)&&(_reaction_2.getError() == INCORRECT_REACTION))
        {
            // Message from reaction 1
            const ::cpptesk::hw::Message *_message1Spec = _reaction_1.getSpecMessage();
            const ::cpptesk::hw::Message *_message1Impl = _reaction_1.getImplMessage();
            // Message from reaction 2
            const ::cpptesk::hw::Message *_message2Spec = _reaction_2.getSpecMessage();
            const ::cpptesk::hw::Message *_message2Impl = _reaction_2.getImplMessage();
    
            if(_message1Spec->size() == _message2Spec->size())
            {
                _match = FUZZY_MATCH;
                
             //   CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "Reaction number in the vector: ");
                
                // Check base match reaction # 1
                _match11 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = _message1Spec->begin(), j = _message1Impl->begin();
                    i != _message1Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message1Spec).**i;
                    const Field<> &field_Impl = (*_message1Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match11--;
                       // CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "_match11--");
                    }
                }
                
                // Check base match reaction # 2
                _match22 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = _message2Spec->begin(), j = _message2Impl->begin();
                    i != _message2Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message2Spec).**i;
                    const Field<> &field_Impl = (*_message2Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match22--;
                       // CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "_match22--");
                    }
                }
                
                // Check match reaction # 1 Impl & reaction # 2 Spec
                _match12 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = _message2Spec->begin(), j = _message1Impl->begin();
                    i != _message2Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message2Spec).**i;
                    const Field<> &field_Impl = (*_message1Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match12--;
                    //CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "_match12--");

                    }
                }
                
                
                // Check match reaction # 2 Impl & reaction # 1 Spec
                _match21 = FUZZY_MATCH;
                for(::cpptesk::hw::Message::const_iterator i = _message1Spec->begin(), j = _message2Impl->begin();
                    i != _message1Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message1Spec).**i;
                    const Field<> &field_Impl = (*_message2Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match21--;
                        //CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "_match11--");
                    }
                }
                
                if(((_match21 < _match11)&&(_match21 < _match22))||((_match12 < _match11)&&(_match12 < _match22)))
                {
                   // CPPTESK_DEBUG_PRINT(DEBUG_ERROR, "_match12 < _match21");
                    if(_match12 < _match21)
                    {
                        _match = _match12;
                    }
                    else
                    {
                        _match = _match21;
                    }
                }
            }
        }
        
        return _match;
    }

    
}}}
