/*Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Fuzzy Impl Intersection Elimination 
//           (Rule # 11 in C++TESK.ErrorDiagnostics.ru.doc):
//           {[Y1:x]+[Z1:Y2] => [Z1:x]+[Y1:Y2]}
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/rule/rule.hpp>
#include <hw/diagnostics/rule/fuzzy_impl_intersection.hpp>
#include <hw/diagnostics/transform_unit.hpp>
#include <hw/diagnostics/reaction_set.hpp>

#include <stdio.h>
#include <hw/message.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

    ReactionSet FuzzyImplIntersectionEliminationRule::transform(::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_1, ::cpptesk::hw::diagnostics::ReactionUnit &_reactionUnit_2)    // Collapse the reactions
    {
        ReactionSet _transform_out;

        // Get "start" and "end" time of reactions
        cycle_t start_time   =  _reactionUnit_1._reaction.whenStarted();
        cycle_t start_time_2 =  _reactionUnit_2._reaction.whenStarted();
        cycle_t end_time     =  _reactionUnit_1._reaction.whenStopped();
        cycle_t end_time_2   =  _reactionUnit_2._reaction.whenStopped();
        
        if(start_time > start_time_2) start_time = start_time_2; // Min start time
        if(end_time < end_time_2)     end_time = end_time_2;     // Max end time
        
        if(_reactionUnit_1._reaction.getError() == UNEXPECTED_REACTION)
        {
            const ::cpptesk::hw::Message *_message1Impl;
            const ::cpptesk::hw::Message *_message2Impl, *_message2Spec;
            
            // Message from reaction 1 [UNEXPECTED] 
            _message1Impl = _reactionUnit_1._reaction.getImplMessage();
            // Message from reaction 2 [INCORRECT]
            _message2Spec = _reactionUnit_2._reaction.getSpecMessage();
            _message2Impl = _reactionUnit_2._reaction.getImplMessage();
            
            const ::cpptesk::hw::Interface *iface1 = _reactionUnit_1._reaction.where();
            const ::cpptesk::hw::Interface *iface2 = _reactionUnit_2._reaction.where();

            ::cpptesk::hw::Reaction reactionIncorrect = Reaction(INCORRECT_REACTION, *iface2, *_message2Spec, *_message1Impl, start_time, end_time, "Incorrect reaction");
            ::cpptesk::hw::Reaction reactionUnexpected = Reaction(UNEXPECTED_REACTION, *iface1, *_message2Impl, start_time, end_time, "Unexpected reaction");

            _transform_out.addReaction(reactionIncorrect);
            _transform_out.addReaction(reactionUnexpected);
        }
        else
        {
            const ::cpptesk::hw::Message *_message1Impl, *_message1Spec;
            const ::cpptesk::hw::Message *_message2Impl;
            
            // Message from reaction 1 [INCORRECT]
            _message1Spec = _reactionUnit_1._reaction.getSpecMessage();
            _message1Impl = _reactionUnit_1._reaction.getImplMessage();
            // Message from reaction 1 [UNEXPECTED] 
            _message2Impl = _reactionUnit_2._reaction.getImplMessage();
            
            const ::cpptesk::hw::Interface *iface1 = _reactionUnit_1._reaction.where();
            const ::cpptesk::hw::Interface *iface2 = _reactionUnit_2._reaction.where();
            
            ::cpptesk::hw::Reaction reactionIncorrect = Reaction(INCORRECT_REACTION, *iface1, *_message1Spec, *_message2Impl, start_time, end_time, "Incorrect reaction");
            ::cpptesk::hw::Reaction reactionUnexpected = Reaction(UNEXPECTED_REACTION, *iface2, *_message1Impl, start_time, end_time, "Unexpected reaction");
            
            _transform_out.addReaction(reactionIncorrect);
            _transform_out.addReaction(reactionUnexpected);
        }
        
        // Set flag
        _reactionUnit_1.useInCollapse();
        _reactionUnit_2.useInCollapse();
       
        return _transform_out;
    }

    int FuzzyImplIntersectionEliminationRule::match(::cpptesk::hw::Reaction _reaction_1, ::cpptesk::hw::Reaction _reaction_2)
    // Compare the reactions
    {
        int _match = FALSE_MATCH, _match1 = FALSE_MATCH, _match2 = FALSE_MATCH;

        if((_reaction_1.getError() == INCORRECT_REACTION)&&(_reaction_2.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            // Get Impl & Spec from incorrect reaction 1
            const ::cpptesk::hw::Message *_message1Spec = _reaction_1.getSpecMessage();
            const ::cpptesk::hw::Message *_message1Impl = _reaction_1.getImplMessage();
            // Message from reaction 2
            // Get Impl from unexpected reaction 2
            const ::cpptesk::hw::Message *_message2Impl = _reaction_2.getImplMessage();
    
            if(_message1Spec->size() == _message2Impl->size())
            {
                _match = FUZZY_MATCH;
                
                // Check match [reaction#1 Impl] & [reaction#1 Spec]
                for(::cpptesk::hw::Message::const_iterator i = _message1Spec->begin(), j = _message1Impl->begin();
                    i != _message1Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message1Spec).**i;
                    const Field<> &field_Impl = (*_message1Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match1--;
                    }
                }

                // Check match [reaction#2 Impl] & [reaction#1 Spec]
                for(::cpptesk::hw::Message::const_iterator i = _message1Spec->begin(), j = _message2Impl->begin();
                    i != _message1Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message1Spec).**i;
                    const Field<> &field_Impl = (*_message2Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match2--;
                    }
                }
                
                if(_match1>_match2)
                {
                    _match = _match1;
                }
                
                if(_match1<_match2)
                {
                    _match = _match2;
                }
                // if _match1 = _match2, then _match = FUZZY_MATCH
            }
        }

        if((_reaction_2.getError() == INCORRECT_REACTION)&&(_reaction_1.getError() == UNEXPECTED_REACTION))
        {
            // Message from reaction 1
            // Get Impl from unexpected reaction 1
            const ::cpptesk::hw::Message *_message1Impl = _reaction_1.getImplMessage();
            // Message from reaction 2
            // Get Impl & Spec from incorrect reaction 2
            const ::cpptesk::hw::Message *_message2Spec = _reaction_2.getSpecMessage();
            const ::cpptesk::hw::Message *_message2Impl = _reaction_2.getImplMessage();
    
            if(_message1Impl->size() == _message2Impl->size())
            {
                _match = FUZZY_MATCH;
                
                // Check match [reaction#1 Impl] & [reaction#2 Spec]
                for(::cpptesk::hw::Message::const_iterator i = _message2Spec->begin(), j = _message1Impl->begin();
                    i != _message2Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message2Spec).**i;
                    const Field<> &field_Impl = (*_message1Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match1--;
                    }
                }

                // Check match [reaction#2 Impl] & [reaction#2 Spec]
                for(::cpptesk::hw::Message::const_iterator i = _message2Spec->begin(), j = _message2Impl->begin();
                    i != _message2Spec->end(); i++, j++)
                {
                    const Field<> &field_Spec = (*_message2Spec).**i;
                    const Field<> &field_Impl = (*_message2Impl).**j;

                    if(field_Spec.equals(field_Impl))
                    {
                        _match2--;
                    }
                }
                
                if(_match1>_match2)
                {
                    _match = _match1;
                }
                
                if(_match1<_match2)
                {
                    _match = _match2;
                }
                // if _match1 = _match2, then _match = FUZZY_MATCH
            }
        }
        
        return _match;
    }

    
}}}
